/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        item.cpp ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include "item.h"
#include "gamedef.h"
#include "mob.h"
#include "defedit.h"

#include <assert.h>

#include <iostream>
using namespace std;

#include "grammar.h"
#include "text.h"
#include "engine.h"

int glbUniqueId = 0;

static PTRLIST<bool> glbItemIded;
static PTRLIST<int> glbItemMagicClass;

int 
glb_allocUID()
{
    return glbUniqueId++;
}

void
glb_reportUID(int uid)
{
    if (uid >= glbUniqueId)
	glbUniqueId = uid+1;
}


//
// Item Definitions
//

ITEM::ITEM()
{
    myDefinition = (ITEM_NAMES) 0;
    myCount = 1;
    myTimer = -1;
    myUID = INVALID_UID;
    myInterestedMobUID = INVALID_UID;
    myBroken = false;
    myEquipped = false;
    myMobType = MOB_NONE;
    myElement = ELEMENT_NONE;
}

ITEM::~ITEM()
{
    pos().removeItem(this);
}

void
ITEM::move(POS pos)
{
    myPos.removeItem(this);
    myPos = pos;
    myPos.addItem(this);
}

void
ITEM::clearAllPos()
{
    myPos = POS();
}

ITEM *
ITEM::copy() const
{
    ITEM *item;

    item = new ITEM();

    *item = *this;

    return item;
}

ITEM *
ITEM::createCopy() const
{
    ITEM *item;

    item = copy();

    item->myUID = glb_allocUID();

    return item;
}

ITEM *
ITEM::create(ITEM_NAMES item, int depth)
{
    ITEM	*i;

    assert(item >= ITEM_NONE && item < GAMEDEF::getNumItem());

    i = new ITEM();

    i->myDefinition = item;
    i->myTimer = i->defn().timer;
    i->setStackCount(i->defn().startstack);

    i->myUID = glb_allocUID();

    if (i->defn().startsbroken)
	i->myBroken = true;

    return i;
}

ITEM *
ITEM::createRandomBalanced(int depth)
{
    // Decide on type
    ITEMCLASS_NAMES	choices[] =
    {
	ITEMCLASS_MELEEWEAPON,
	ITEMCLASS_ARMOUR,
	ITEMCLASS_POTION,
	ITEMCLASS_SPELLBOOK,
    };
    ITEMCLASS_NAMES 	itemclass = choices[rand_choice(4)];
    ITEM	*item = 0;

    // Quit early if not unlocked.
    int		totaldeaths = GAMEDEF::mobdef(MOB_AVATAR)->totalkillsever;

    if (totaldeaths < GAMEDEF::itemclassdef(itemclass)->neededdeaths)
	return 0;

    int		matchingdepth = 0;
    int		choice = 0;
    ITEM_NAMES	itemname = ITEM_NONE;

    for (ITEM_NAMES i = ITEM_NONE; i < (ITEM_NAMES) GAMEDEF::getNumItem(); i = (ITEM_NAMES) (i+1))
    {
	// Ignore mismatching.
	if (!defn(i).depth)
	    continue;

	if (defn(i).itemclass != itemclass)
	    continue;

	if (defn(i).depth <= depth)
	{
	    if (defn(i).depth == depth)
		matchingdepth++;

	    if (rand_choice(choice + defn(i).rarity) < defn(i).rarity)
		itemname = (ITEM_NAMES) i;
	    choice += defn(i).rarity;
	}
    }

    // See if we failed..
    if (!matchingdepth || 
	(matchingdepth < 3 && !rand_choice(matchingdepth*20)))
    {
	// No mobs created on this level, or we have gotten bored
	// our prebuilt ones.
	itemname = GAMEDEF::createNewItemDef();

	GAMEDEF::itemdef(itemname)->depth = depth;
	GAMEDEF::itemdef(itemname)->itemclass = itemclass;

	GAMEDEF::itemdef(itemname)->symbol = GAMEDEF::itemclassdef(itemclass)->symbol;

	if (itemclass == ITEMCLASS_SPELLBOOK)
	{
	    // We need a spell to go with this.
	    SPELL_NAMES		spell = GAMEDEF::createNewSpellDef();

	    // Cross link the two.
	    GAMEDEF::itemdef(itemname)->spell = spell;
	    GAMEDEF::spelldef(spell)->item = itemname;
	    GAMEDEF::spelldef(spell)->mana = MOB::expectedMPforLevel(depth) / 5;
	}
    }

    return create(itemname, depth);
}

ITEM *
ITEM::createRandom(int depth)
{
    ITEM_NAMES		item;

    item = itemFromHash(rand_choice(65536*32767), depth);
    if (item == ITEM_NONE)
	return 0;
    return create(item, depth);
}

ITEM_NAMES
ITEM::itemFromHash(unsigned int hash, int depth)
{
    int		totalrarity, rarity;
    int		i;

    // Find total rarity of items
    totalrarity = 0;
    for (i = ITEM_NONE; i < GAMEDEF::getNumItem(); i++)
    {
	rarity = defn((ITEM_NAMES)i).rarity;
	if (defn((ITEM_NAMES)i).depth > depth)
	    rarity = 0;
	totalrarity += rarity;
    }

    hash %= totalrarity;
    for (i = ITEM_NONE; i < GAMEDEF::getNumItem(); i++)
    {
	rarity = defn((ITEM_NAMES)i).rarity;
	if (defn((ITEM_NAMES)i).depth > depth)
	    rarity = 0;

	if (hash > (unsigned) rarity)
	    hash -= rarity;
	else
	    break;
    }
    return (ITEM_NAMES) i;
}

int
ITEM::getMagicClass() const
{
    return glbItemMagicClass.safeIndexAnywhere(getDefinition());
}

bool
ITEM::isMagicClassKnown() const
{
    return glbItemIded.safeIndexAnywhere(getDefinition());
}

void
ITEM::markMagicClassKnown()
{
    glbItemIded.indexAnywhere(getDefinition()) = true;
}

VERB_PERSON
ITEM::getPerson() const
{
    return VERB_IT;
}

BUF
ITEM::getSingleName() const
{
    BUF		buf, basename;

    basename = getRawName();

    if (mobType() != MOB_NONE)
    {
	const char *mobname = GAMEDEF::mobdef(mobType())->name;
	buf.sprintf("%s %s", mobname, basename.buffer());
	basename = buf;
    }
    if (element() != ELEMENT_NONE)
    {
	const char *elementname = GAMEDEF::elementdef(element())->name;
	buf.sprintf("%s %s", basename.buffer(), elementname);
	basename = buf;
    }

    if (!buf.isstring())
    {
	if (getTimer() >= 0)
	{
	    // A timed item...
	    buf.sprintf("%s (%d)", basename.buffer(), getTimer());
	}
	else
	{
	    // A normal item.
	    buf.reference(basename.buffer());
	}
    }
    if (isBroken())
    {
	BUF		broken;
	broken.sprintf("broken %s", buf.buffer());
	buf.strcpy(broken);
    }

    return buf;
}

BUF
ITEM::getName() const
{
    BUF		buf = getSingleName();
    return gram_createcount(buf, myCount, false);
}

BUF
ITEM::getArticleName() const
{
    BUF		name = getName();
    BUF		result;

    result.sprintf("%s%s", gram_getarticle(name.buffer()), name.buffer());

    return result;
}

BUF
ITEM::getRawName() const
{
    BUF		buf;

    buf.reference(defn().name);


    if (getDefinition() == ITEM_CORPSE)
    {
	if (MOB::defn(mobType()).isfriendly)
	    buf.reference("body");
    }

    if (isPotion())
    {
	if (isMagicClassKnown())
	{
	    buf.reference(glb_potiondefs[getMagicClass()].name);
	}
    }
    if (isRing())
    {
	if (isMagicClassKnown())
	{
	    buf.reference(glb_ringdefs[getMagicClass()].name);
	}
    }

    return buf;
}

BUF
ITEM::getDetailedDescription() const
{
    BUF		result, buf;

    result.reference("");

    if (isMelee())
    {
	buf.sprintf(
		    "Average Damage: %d\n"
		    "Distribution: %s\n"
		    "To Hit Bonus: %d\n"
		    "Element: %s\n",
		    attackdefn().damage,
		    glb_distributiondefs[attackdefn().distribution].name,
		    attackdefn().chancebonus,
		    GAMEDEF::elementdef(attackdefn().element)->name
		    );
	result.strcat(buf);
    }
    if (isRanged())
    {
    }
    if (isArmour())
    {
	buf.sprintf(
		    "Damage Reduction: %d\n",
		    getDamageReduction());
	result.strcat(buf);
    }

    if (isRing() && isMagicClassKnown())
    {
	RING_NAMES	ring = (RING_NAMES) getMagicClass();

	if (glb_ringdefs[ring].resist != ELEMENT_NONE)
	{
	    BUF		capelement;
	    capelement = gram_capitalize(GAMEDEF::elementdef(glb_ringdefs[ring].resist)->name);
	    int		ramt = glb_ringdefs[ring].resist_amt;
	    if (ramt < 0)
		buf.sprintf("%s Vulnerable: %d\n", capelement.buffer(), -ramt);
	    else
		buf.sprintf("%s Resist: %d\n", capelement.buffer(), ramt);
	    result.strcat(buf);
	}
	if (glb_ringdefs[ring].deflect != 0)
	{
	    int		amt = glb_ringdefs[ring].deflect;
	    buf.sprintf("Deflection: %d\n", amt);
	    result.strcat(buf);
	}
    }
    return result;
}

BUF
ITEM::getLongDescription() const
{
    BUF		descr, detail;

    descr = gram_capitalize(getName());
    detail = getDetailedDescription();
    descr.append('\n');
    if (detail.isstring())
    {
	descr.append('\n');
	descr.strcat(detail);
    }

    detail = text_lookup("item", getRawName());
    if (detail.isstring() && !detail.startsWith("Missing text entry: "))
    {
	descr.append('\n');
	descr.strcat(detail);
    }
    else
    {
	descr.append('\n');
    }

    // Append the non-magic version
    if (isMagicClassKnown() && (isPotion() || isRing()))
    {
	descr.strcat("Base Type: ");
	detail = gram_capitalize(defn().name);
	detail.append('\n');
	descr.strcat(detail);

	detail = text_lookup("item", defn().name);
	if (detail.isstring() && !detail.startsWith("Missing text entry: "))
	{
	    descr.append('\n');
	    descr.strcat(detail);
	}
	else
	{
	    descr.append('\n');
	}
    }

    return descr;
}

void
ITEM::getLook(u8 &symbol, ATTR_NAMES &attr) const
{
    symbol = defn().symbol;
    attr = (ATTR_NAMES) defn().attr;
}

bool
ITEM::canStackWith(const ITEM *stack) const
{
    if (getDefinition() != stack->getDefinition())
	return false;
    
    if (mobType() != stack->mobType())
	return false;

    if (isBroken() != stack->isBroken())
	return false;

    // Disable stacking of weapons/books/armour as we want
    // to be able to delete them easily
    if (defn().unstackable)
	return false;

    // No reason why not...
    return true;
}

void
ITEM::combineItem(const ITEM *item)
{
    // Untimed items stack.
    if (getTimer() < 0)
	myCount += item->getStackCount();
    // Timed items add up charges.
    if (getTimer() >= 0)
    {
	assert(item->getTimer() >= 0);
	myTimer += item->getTimer();
    }

    assert(myCount >= 0);
    if (myCount < 0)
	myCount = 0;
}

bool
ITEM::runHeartbeat()
{
    if (myTimer >= 0)
    {
	if (!myTimer)
	    return true;
	myTimer--;
    }
    return false;
}

void
ITEM::initSystem()
{
    glbItemIded.clear();
    glbItemMagicClass.clear();

    // Mix the rings...
    int		ringclasses[NUM_RINGS];
    RING_NAMES	ring;

    int pclass = 1;
    for (ITEM_NAMES item = ITEM_NONE; item < GAMEDEF::getNumItem();
	 item = (ITEM_NAMES)(item+1))
    {
	if (defn(item).itemclass == ITEMCLASS_RING)
	{
	    if (pclass >= NUM_RINGS)
	    {
		assert(!"Too many mundane rings");
		break;
	    }
	    ringclasses[pclass] = item;
	    pclass++;
	}
    }
    rand_shuffle(&ringclasses[1], pclass-1);
    FOREACH_RING(ring)
    {
	if (ring == RING_NONE)
	    continue;
	if (ring >= pclass)
	{
	    assert(!"Not enough mundane rings");
	    break;
	}
	glbItemMagicClass.indexAnywhere(ringclasses[ring]) = ring;
    }

}

void
ITEM::saveGlobal(ostream &os)
{
    int		val;

    val = GAMEDEF::getNumItem();
    os.write((const char *)&val, sizeof(int));

    for (ITEM_NAMES item = ITEM_NONE; item < GAMEDEF::getNumItem();
	 item = (ITEM_NAMES)(item+1))
    {
	val = glbItemIded.safeIndexAnywhere(item);
	os.write((const char *) &val, sizeof(int));
	val = glbItemMagicClass.safeIndexAnywhere(item);
	os.write((const char *) &val, sizeof(int));
    }
}

void
ITEM::save(ostream &os) const
{
    int		val;

    val = myDefinition;
    os.write((const char *) &val, sizeof(int));

    myPos.save(os);

    os.write((const char *) &myCount, sizeof(int));
    os.write((const char *) &myTimer, sizeof(int));
    os.write((const char *) &myUID, sizeof(int));
    os.write((const char *) &myInterestedMobUID, sizeof(int));
    os.write((const char *) &myBroken, sizeof(bool));
    os.write((const char *) &myEquipped, sizeof(bool));
    val = myMobType;
    os.write((const char *) &val, sizeof(int));
    val = myElement;
    os.write((const char *) &val, sizeof(int));
}

ITEM *
ITEM::load(istream &is)
{
    int		val;
    ITEM	*i;

    i = new ITEM();

    is.read((char *)&val, sizeof(int));
    i->myDefinition = (ITEM_NAMES) val;

    i->myPos.load(is);

    is.read((char *)&i->myCount, sizeof(int));
    is.read((char *)&i->myTimer, sizeof(int));

    is.read((char *)&i->myUID, sizeof(int));
    glb_reportUID(i->myUID);
    is.read((char *)&i->myInterestedMobUID, sizeof(int));
    is.read((char *)&i->myBroken, sizeof(bool));
    is.read((char *)&i->myEquipped, sizeof(bool));
    is.read((char *)&val, sizeof(int));
    i->myMobType = (MOB_NAMES) val;
    is.read((char *)&val, sizeof(int));
    i->myElement = (ELEMENT_NAMES) val;

    return i;
}

void
ITEM::loadGlobal(istream &is)
{
    int		val, total;

    is.read((char *) &total, sizeof(int));

    for (int i = 0; i < total; i++)
    {
	is.read((char *) &val, sizeof(int));
	glbItemIded.indexAnywhere(i) = val ? true : false;
	is.read((char *) &val, sizeof(int));
	glbItemMagicClass.indexAnywhere(i) = val;
    }
}

bool
ITEM::isPotion() const
{
    if (defn().itemclass == ITEMCLASS_POTION)
	return true;
    return false;
}

bool
ITEM::isSpellbook() const
{
    if (defn().itemclass == ITEMCLASS_SPELLBOOK)
	return true;
    return false;
}

bool
ITEM::isRing() const
{
    if (defn().itemclass == ITEMCLASS_RING)
	return true;
    return false;
}

bool
ITEM::isFood() const
{
    if (defn().itemclass == ITEMCLASS_FOOD)
	return true;
    return false;
}

bool
ITEM::isMelee() const
{
    if (defn().itemclass == ITEMCLASS_MELEEWEAPON)
	return true;
    return false;
}

bool
ITEM::isArmour() const
{
    if (defn().itemclass == ITEMCLASS_ARMOUR)
	return true;
    return false;
}

bool
ITEM::isRanged() const
{
    if (defn().itemclass == ITEMCLASS_RANGEWEAPON)
	return true;
    return false;
}

void
ITEM::getRangeStats(int &range, int &power, int &consistency, int &area) const
{
    range = defn().range_range;
    consistency = defn().range_consistency;
    power = defn().range_power;
    area = defn().range_area;
}

int
ITEM::getRangeRange() const
{
    // Compute our power, accuracy, and consistency.
    int		range, d;

    getRangeStats(range, d, d, d);

    return range;
}

int
ITEM::getRangeArea() const
{
    // Compute our power, accuracy, and consistency.
    int		area, d;

    getRangeStats(d, d, d, area);

    return area;
}

void
ITEM::setupTextVariables() const
{
    BUF		tmp;

    tmp.strcpy(gram_getarticle(getName()));
    tmp.strcat(getName());
    text_registervariable("A_ITEMNAME", tmp);
    text_registervariable("ITEMNAME", getName());
    text_registervariable("ITEMNAMES", gram_makepossessive(getName()));
    tmp.sprintf("%d", defn().depth);
    text_registervariable("ITEMDEPTH", tmp);
    text_registervariable("ITEMCLASS", GAMEDEF::itemclassdef(defn().itemclass)->name);
}

void
ITEM::ensureNamed() const
{
    if (!strcmp(defn().name, "unnamed"))
    {
	setupTextVariables();
	const char *query = "item_name";
	BUF	newname = glbEngine->askQuestionTextCanned("query", query, false);
	if (newname.isstring())
	{
	    edefn().name = glb_harden(newname);
	    // WE don't update the symbol as traditionally these are left
	    // to the item class.
	}
    }
}

void
ITEM::ensurePotionEffect(ITEM_NAMES item)
{
    if (defn(item).potion_effect == EFFECT_NONE)
    {
	edefn(item).potion_effect = defedit_editEffect((EFFECT_NAMES) defn(item).potion_effect, defn(item).depth, false);
    }
}

EFFECT_NAMES
ITEM::corpseEffect() const
{
    if (getDefinition() != ITEM_CORPSE)
	return EFFECT_NONE;

    if (mobType() == MOB_NONE)
	return EFFECT_NONE;

    return (EFFECT_NAMES) MOB::defn(mobType()).corpse_effect;
}

void
ITEM::ensureCorpseEffect() const
{
    if (getDefinition() != ITEM_CORPSE)
	return;

    if (mobType() == MOB_NONE)
	return;

    if (corpseEffect() != EFFECT_NONE)
	return;

    if (glbEngine->timeForQuestions() &&
	MOB::defn(mobType()).totalkillsever >= 6)
    {
	BUF		prequel;
	BUF		name = getName();
	prequel.sprintf("What extra thing happens when you eat %s?", name.buffer());

	MOB::edefn(mobType()).corpse_effect =
	    defedit_editEffect(prequel, 
				(EFFECT_NAMES)MOB::defn(mobType()).corpse_effect,
				MOB::defn(mobType()).depth,
				false);
    }
}

