/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        mob.cpp ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include "mob.h"

#include "map.h"
#include "msg.h"
#include "text.h"
#include "item.h"
#include "event.h"
#include "display.h"
#include "engine.h"
#include "simulator.h"
#include "defedit.h"

#include <stdio.h>

#include <iostream>
using namespace std;

//
// Fireball operators
//
class ATTACK_OP
{
public:
    ATTACK_OP(MOB *src, ATTACK_NAMES attack)
    {
	mySrc = src;
	myAttack = attack;
    }

    void operator()(POS p)
    {
	if (p.mob() && p.mob() != mySrc)
	{
	    if (mySrc->evaluateWillHit(myAttack, p.mob()))
	    {
		if (!p.mob()->applyDamage(mySrc, 
			    mySrc->evaluateAttackDamage(myAttack),
			    (ELEMENT_NAMES) GAMEDEF::attackdef(myAttack)->element,
			    ATTACKSTYLE_RANGE))
		{
		    // Mob still alive.
		    if (p.mob() && GAMEDEF::attackdef(myAttack)->effect != EFFECT_NONE)
		    {
			p.mob()->applyEffect(mySrc,
			    (EFFECT_NAMES) GAMEDEF::attackdef(myAttack)->effect,
			    ATTACKSTYLE_RANGE);
		    }
		}
	    }
	    else
	    {
		mySrc->formatAndReport("%s <miss> %O.", p.mob());
	    }
	}
    }

private:
    ATTACK_NAMES	 myAttack;
    MOB			*mySrc;
};

class POTION_OP
{
public:
    POTION_OP(MOB *src, ITEM_NAMES potion, bool *interest) 
    { mySrc = src; myPotion = potion; myInterest = interest; }

    void operator()(POS p)
    {
	MOB		*mob = p.mob();

	if (mob)
	{
	    ITEM::ensurePotionEffect(myPotion);
	    mob->applyEffect(mySrc, (EFFECT_NAMES) GAMEDEF::itemdef(myPotion)->potion_effect,
		ATTACKSTYLE_RANGE);
	}
    }

private:
    bool		*myInterest;
    ITEM_NAMES	 	myPotion;
    MOB			*mySrc;
};


//
// MOB Implementation
//

MOB::MOB()
{
    myFleeCount = 0;
    myBoredom = 0;
    myYellHystersis = 0;
    myAIState = 0;
    myHP = 0;
    // Enter the dungeon fully fed!
    myFood = getMaxFood();
    myLevel = 0;
    myIsSwallowed = false;
    myNumDeaths = 0;
    myUID = INVALID_UID;
    mySkipNextTurn = false;
    myDelayMob = false;
    myDelayMobIdx = -1;
    myCollisionTarget = 0;
    mySearchPower = 0;
    myAngryWithAvatar = false;

    YELL_NAMES	yell;
    FOREACH_YELL(yell)
    {
	myHeardYell[yell] = false;
	myHeardYellSameRoom[yell] = false;
    }
    mySawMurder = false;
    mySawMeanMurder = false;
    mySawVictory = false;
    myAvatarHasRanged = false;
    myWaiting = false;
    myCowardice = 0;
    myKills = 0;
    myHasWon = false;
    myRangeTimeout = 0;
}

MOB::~MOB()
{
    int		i;

    if (myPos.map() && myPos.map()->avatar() == this)
	myPos.map()->setAvatar(0);

    myPos.removeMob(this);

    for (i = 0; i < myInventory.entries(); i++)
	delete myInventory(i);
}

MOB *
MOB::create(MOB_NAMES def)
{
    MOB		*mob;

    mob = new MOB();

    mob->myDefinition = def;

    mob->myHP = defn(def).max_hp;
    mob->myMP = defn(def).max_mp;
    mob->myLevel = defn(def).depth;

    mob->myUID = glb_allocUID();

    return mob;
}

MOB *
MOB::copy() const
{
    MOB		*mob;

    mob = new MOB();
    
    *mob = *this;

    // Copy inventory one item at a time.
    // We are careful to maintain the same list order here so restoring
    // won't shuffle things unexpectadly
    int		 i;

    mob->myInventory.clear();
    for (i = 0; i < myInventory.entries(); i++)
    {
	mob->myInventory.append(myInventory(i)->copy());
    }
    
    return mob;
}

void
MOB::setPos(POS pos)
{
    myPos.removeMob(this);
    myPos = pos;
    myPos.addMob(this);
}


void
MOB::setMap(MAP *map)
{
    myPos.setMap(map);
    myTarget.setMap(map);
    myHome.setMap(map);
    myMeditatePos.setMap(map);
}

void
MOB::clearAllPos()
{
    myPos = POS();
    myTarget = POS();
    myHome = POS();
    myMeditatePos = POS();
}

MOB *
MOB::createNPC(int depth)
{
    int		i;
    MOB_NAMES	mob = MOB_NONE;
    int		choice = 0;
    int		matchingdepth = 0;
    MOB		*m;
    int		maxmobtype = GAMEDEF::mobdef(MOB_AVATAR)->totalkillsever / 5;
    maxmobtype = MIN(maxmobtype, 4);
    maxmobtype = MAX(maxmobtype, 1);

    // Given the letter choice, choose a mob that matches it appropriately.
    choice = 0;
    for (i = 0; i < GAMEDEF::getNumMob(); i++)
    {
	// Stuff with 0 depth is never created manually.
	if (!defn(i).depth)
	    continue;

	// Use the baseletter to bias the creation.
	if (defn(i).depth <= depth)
	{
	    if (defn(i).depth == depth)
		matchingdepth++;
	    if (rand_choice(choice + defn(i).rarity) < defn(i).rarity)
		mob = (MOB_NAMES) i;
	    choice += defn(i).rarity;
	}
    }

    if (!matchingdepth || 
	(matchingdepth < maxmobtype && !rand_choice(matchingdepth*20)))
    {
	// No mobs created on this level, or we have gotten bored
	// our prebuilt ones.
	mob = GAMEDEF::createNewMobDef();

	GAMEDEF::mobdef(mob)->depth = depth;
    }

    if (mob == MOB_NONE)
    {
	// Welp, we create a new one!
	return 0;
    }

    // Testing..
    // mob = MOB_PYTHON;

    m = MOB::create(mob);

    if (0)
    {
	ITEM *item = ITEM::createRandom(depth);
	if (item)
	    m->addItem(item);
    }

    return m;
}

void
MOB::getLook(u8 &symbol, ATTR_NAMES &attr) const
{
    symbol = defn().symbol;
    attr = (ATTR_NAMES) defn().attr;

    if (!alive())
    {
	// Dead creatures override their attribute to blood
	attr = ATTR_RED;
    }
}

const char *
MOB::getName() const
{
    if (isAvatar())
	return "you";

    if (myName.isstring())
	return myName.buffer();

    return defn().name;
}

const char *
MOB::getRawName() const
{
    return defn().name;
}

void
MOB::setupTextVariables() const
{
    BUF		tmp;

    tmp.sprintf("%s%s", gram_getarticle(getName()), getName());
    text_registervariable("A_MOBNAME", tmp);
    text_registervariable("MOBNAME", getName());
    text_registervariable("MOBNAMES", gram_makepossessive(getName()));
    tmp.sprintf("%d", getLevel());
    text_registervariable("MOBDEPTH", tmp);
}


void
MOB::ensureSpell(SPELL_NAMES spell)
{
    if (GAMEDEF::spelldef(spell)->effect == EFFECT_NONE)
    {
	ENGINE::Question		question;
	question.payload = spell;
	question.type = ENGINE::QUERY_SPELL;
	glbEngine->askQuestion(question, false);
    }
}

void
MOB::ensureNamed() const
{
    if (!strcmp(defn().name, "unnamed"))
    {
	setupTextVariables();
	const char *query = "mob_name";
	if (isBoss())
	    query = "boss_name";
	BUF	newname = glbEngine->askQuestionTextCanned("query", query, false);
	if (newname.isstring())
	{
	    GAMEDEF::mobdef(getDefinition())->name = glb_harden(newname);
	    GAMEDEF::mobdef(getDefinition())->symbol = newname.buffer()[0];
	}
    }
}

void
MOB::ensureAI()
{
    if ((isBoss() && !defn().set_ai) ||
	(glbEngine->timeForQuestions() &&
	!defn().set_ai && 
	 defn().totalkillsever >= 5))
    {
	defedit_editMobAI(this, false);
    }
}

void
MOB::ensureRarity()
{
    if (glbEngine->timeForQuestions() &&
	!defn().set_rarity && 
	 defn().totalkillsever >= 12)
    {
	defedit_editMobRarity(this, false);
    }
}

void
MOB::ensureSpecial()
{
    if (glbEngine->timeForQuestions() &&
	!defn().set_special && 
	 defn().totalkillsever >= 10)
    {
	defedit_editMobSpecial(this, false);
    }
}

void
MOB::ensureMeleeAttack()
{
    ITEM		*w = lookupWeapon();

    if (!w)
    {
	ATTACK_NAMES	attack = (ATTACK_NAMES) defn().melee_attack;

	if (attack == ATTACK_NONE)
	{
	    defedit_editAttackDamage(this, 0, attack, false);
	}

	attack = (ATTACK_NAMES) defn().melee_attack;
	if ( (isBoss() && !GAMEDEF::attackdef(attack)->set_special)
	    || (glbEngine->timeForQuestions() && 
		!GAMEDEF::attackdef(attack)->set_special &&
		defn().totalkillsever > 3) )
	{
	    defedit_editAttack(this, 0, attack, false);
	}
    }
    else
    {
	ATTACK_NAMES	attack = (ATTACK_NAMES) w->defn().melee_attack;

	if (attack == ATTACK_NONE)
	{
	    defedit_editAttackDamage(0, w, attack, false);
	}
	else
	{
	    if (glbEngine->timeForQuestions() &&
		!GAMEDEF::attackdef(attack)->set_special &&
		!rand_choice(20))
	    {
		defedit_editAttack(0, w, attack, false);
	    }
	}
    }
}

void
MOB::ensureRange()
{
    if ((isBoss() && !defn().set_range) ||
	(glbEngine->timeForQuestions() &&
	 !defn().set_range &&
	  defn().totalkillsever >= 5))
    {
	defedit_editRangeAttack(this, false);
    }
}

void
MOB::ensureDamageReduction()
{
    if ((isBoss() && !defn().set_damagereduction) ||
	(glbEngine->timeForQuestions() &&
	 !defn().set_damagereduction &&
	  defn().totalkillsever >= 7))
    {
	defedit_editMobArmour(this, false);
    }

    ITEM	*armour = lookupArmour();
    if (armour)
    {
	if (!armour->defn().set_damagereduction)
	{
	    defedit_editItemArmour(armour, false);
	}
    }
}

void
MOB::ensureHitPoints()
{
    if (!defn().set_hp)
    {
	defedit_editMobHP(this, false);
    }
}

void
MOB::ensureCorpseChance()
{
    if (!defn().set_corpsechance)
    {
	defedit_editMobCorpseChance(this, false);
    }
}

void
MOB::ensureLoot()
{
    if (glbEngine->timeForQuestions() &&
	(defn().totalkillsever >= 4) &&
	defn().loot < 0)
    {
	defedit_editMobLoot(this, false);
    }
}

BUF
MOB::buildSpellDescr(SPELL_NAMES spell)
{
    BUF		stats, descr;

    SPELL_DEF		*def = GAMEDEF::spelldef(spell);

    stats.sprintf("Spell: %s (%s)\n", def->name, def->verb);
    descr.strcat(stats);
    stats.sprintf("Mana: %d\n", def->mana);
    descr.strcat(stats);
    stats.sprintf("Radius: %d\n", def->radius);
    descr.strcat(stats);

    if (def->needsdir)
	descr.strcat("Targetted\n");
    else
	descr.strcat("Targets Self\n");

    stats = buildEffectDescr((EFFECT_NAMES) def->effect);
    if (stats.isstring())
    {
	descr.strcat("Effect: ");
	descr.strcat(stats);
	descr.strcat("\n");
    }

    return descr;
}

BUF
MOB::buildAttackDescr(ATTACK_NAMES attack)
{
    BUF		stats, descr;

    ATTACK_DEF		*def = GAMEDEF::attackdef(attack);
    stats.sprintf("Attack: %s (%s)\n", def->noun, def->verb);
    descr.strcat(stats);
    stats.sprintf("   Damage: %d (%s)\n", def->damage, glb_distributiondefs[def->distribution].name);
    descr.strcat(stats);
    if (def->chancebonus)
    {
	stats.sprintf("   Bonus to Hit: %d\n", def->chancebonus);
	descr.strcat(stats);
    }
    if (def->element != ELEMENT_PHYSICAL)
    {
	stats.sprintf("   Element: %s\n", GAMEDEF::elementdef(def->element)->name);
	descr.strcat(stats);
    }
    stats = buildEffectDescr((EFFECT_NAMES) def->effect);
    if (stats.isstring())
    {
	descr.strcat("   Effect: ");
	descr.strcat(stats);
	descr.append('\n');
    }

    return descr;
}

BUF
MOB::buildSpecialDescr(MOB_NAMES mob)
{
    BUF		descr, stats;

    if (defn(mob).resistance != ELEMENT_NONE)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	stats.sprintf("resist %s", GAMEDEF::elementdef(defn(mob).resistance)->name);
	descr.strcat(stats);
    }
    if (defn(mob).vulnerability != ELEMENT_NONE)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	stats.sprintf("vulnerable to %s", GAMEDEF::elementdef(defn(mob).vulnerability)->name);
	descr.strcat(stats);
    }

    if (defn(mob).isslow)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("slow");
    }
    if (defn(mob).isfast)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("fast");
    }
    if (defn(mob).isvampire)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("vampiric");
    }
    if (defn(mob).isregenerate)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("regenerates");
    }
    if (defn(mob).candig)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("digs");
    }
    if (defn(mob).swallows)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("engulfs");
    }
    if (defn(mob).breeder)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("breeds");
    }
    if (defn(mob).isthief)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("steals");
    }
    if (defn(mob).canleap)
    {
	if (descr.isstring())
	    descr.strcat(", ");
	descr.strcat("leaps");
    }

    return descr;
}

BUF
MOB::getLongDescription() const
{
    BUF		descr, detail;
    BUF		stats;

    descr.strcpy(gram_capitalize(getName()));
    stats.sprintf(" - Health: %d / %d\n", getHP(), getMaxHP());
    descr.strcat(stats);
    descr.append('\n');

    if (!isAvatar())
    {
	stats.sprintf("Behaviour: %s; ", glb_aidefs[getAI()].name);
	descr.strcat(stats);

	stats.sprintf("Abundance: %d%%\n", defn().rarity);
	descr.strcat(stats);
    }
    if (getMaxMP())
    {
	stats.sprintf("Mana: %d / %d\n", getMP(), getMaxMP());
	descr.strcat(stats);
    }
    if (isAvatar())
    {
	stats.sprintf("Food: %d / %d\n", getFood(), getMaxFood());
	descr.strcat(stats);
    }

    int		reduction = getDamageReduction();
    if (reduction)
    {
	stats.sprintf("Damage Reduction: %d", reduction);
	descr.strcat(stats);
    }
    if (defn().dodgebonus)
    {
	if (reduction)
	    descr.strcat("; ");
	stats.sprintf("Dodge Bonus: %d", defn().dodgebonus);
	descr.strcat(stats);
    }
    if (defn().dodgebonus || reduction)
	descr.append('\n');


    if (getMeleeAttack() != ATTACK_NONE)
    {
	descr.append('\n');
	stats = buildAttackDescr(getMeleeAttack());
	descr.strcat(stats);
    }

    if (defn().range_valid && defn().range_range)
    {
	stats.sprintf("\nRange: %d; Recharge %d\n", defn().range_range, defn().range_recharge);
	descr.strcat(stats);
	stats = buildAttackDescr((ATTACK_NAMES)defn().range_attack);
	descr.strcat(stats);
    }

    descr.append('\n');

    stats = buildSpecialDescr(getDefinition());
    if (stats.isstring())
    {
	descr.strcat("Special: ");
	descr.strcat(stats);
	descr.append('\n');
    }

    return descr;
}

ITEM *
MOB::getRandomItem() const
{
    ITEM	*result;
    int		choice = 0;

    result = 0;

    if (myInventory.entries())
    {
	choice = rand_choice(myInventory.entries());
	return myInventory(choice);
    }
    return result;
}

bool
MOB::hasItem(ITEM_NAMES itemname) const
{
    if (lookupItem(itemname))
	return true;

    return false;
}

ITEM *
MOB::giftItem(ITEM_NAMES itemname)
{
    ITEM		*item;

    item = ITEM::create(itemname);

    addItem(item);

    // Allow for stacking!
    item = lookupItem(itemname);

    return item;
}

ITEM *
MOB::lookupWeapon() const
{
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->isMelee() && myInventory(i)->isEquipped())
	    return myInventory(i);
    }
    return 0;
}

int
MOB::getDamageReduction() const
{
    int		reduction = defn().damagereduction;
    ITEM	*armour = lookupArmour();
    if (armour)
    {
	reduction += armour->getDamageReduction();
    }

    return reduction;
}

ITEM *
MOB::lookupArmour() const
{
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->isArmour() && myInventory(i)->isEquipped())
	    return myInventory(i);
    }
    return 0;
}

ITEM *
MOB::lookupWand() const
{
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->isRanged() && myInventory(i)->isEquipped())
	    return myInventory(i);
    }
    return 0;
}

ITEM *
MOB::lookupItem(ITEM_NAMES itemname) const
{
    int		i;

    for (i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->getDefinition() == itemname)
	    return myInventory(i);
    }
    return 0;
}

ITEM *
MOB::lookupRing() const
{
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->isRing() && myInventory(i)->isEquipped())
	    return myInventory(i);
    }
    return 0;
}

RING_NAMES
MOB::lookupRingName() const
{
    ITEM		*ring = lookupRing();
    if (!ring)
	return RING_NONE;

    return (RING_NAMES) ring->getMagicClass();
}

bool
MOB::canMoveDir(int dx, int dy, bool checkmob) const
{
    POS		goal;

    goal = myPos.delta(dx, dy);

    return canMove(goal, checkmob);
}

bool
MOB::canMove(POS pos, bool checkmob) const
{
    if (!pos.valid())
	return false;

    if (!pos.isPassable())
    {
	if (pos.defn().isphaseable && defn().passwall)
	{
	    // Can move through it...
	}
	else if (pos.defn().isdiggable && defn().candig)
	{
	    // Could dig through it.
	}
	else
	{
	    // Failed...
	    return false;
	}
    }

    if (checkmob && pos.mob())
	return false;

    return true;
}

void
MOB::move(POS newpos, bool ignoreangle)
{
    // If we have swallowed something, move it too.
    if (!isSwallowed())
    {
	PTRLIST<MOB *>	moblist;
	int		i;

	pos().getAllMobs(moblist);
	for (i = 0; i < moblist.entries(); i++)
	{
	    if (moblist(i)->isSwallowed())
	    {
		moblist(i)->move(newpos, true);
	    }
	}
    }

    if (ignoreangle)
	newpos.setAngle(pos().angle());

    setPos(newpos);

    reportSquare(pos());
}

int
MOB::getFood() const
{
    return myFood;
}

void
MOB::gainFood(int food)
{
    myFood += food;
    // We allow ourselves to be oversatiated!
}

void
MOB::eatFood(int food)
{
    myFood -= food;
    if (myFood < 0)
    {
	myFood = 0;
    }
}

int
MOB::getMaxFood() const
{
    return 1000;
}

int
MOB::getMaxHP() const
{
    if (isAvatar())
    {
	return expectedHPforLevel(getLevel());
    }
    else
    {
	return defn().max_hp;
    }
}

void
MOB::gainHP(int hp)
{
    int		maxhp;

    maxhp = getMaxHP();

    myHP += hp;
    if (myHP > maxhp)
	myHP = maxhp;
}

int
MOB::getMaxMP() const
{
    if (isAvatar())
    {
	return expectedMPforLevel(getLevel());
    }
    else
    {
	return defn().max_mp;
    }
}

void
MOB::gainMP(int mp)
{
    int		maxmp;

    maxmp = getMaxMP();

    myMP += mp;
    if (myMP > maxmp)
	myMP = maxmp;
}

void
MOB::postEvent(EVENTTYPE_NAMES type, u8 sym, ATTR_NAMES attr, const char *text) const
{
    if (pos().map())
	pos().map()->getDisplay()->queue().append(EVENT((MOB *)this, sym, attr, type, text));
}

void
MOB::doEmote(const char *text)
{
    postEvent( (EVENTTYPE_NAMES) (EVENTTYPE_SHOUT | EVENTTYPE_LONG),
		    ' ', ATTR_EMOTE,
		    text);
}

void
MOB::doShout(const char *text)
{
    postEvent( (EVENTTYPE_NAMES) (EVENTTYPE_SHOUT | EVENTTYPE_LONG),
		    ' ', ATTR_SHOUT,
		    text);
}

BUF
MOB::buildEffectDescr(EFFECT_NAMES effect)
{
    BUF		descr;
    EFFECT_DEF	*def = GAMEDEF::effectdef(effect);

    if (effect == EFFECT_NONE)
	return descr;
    switch ((EFFECTCLASS_NAMES) def->type)
    {
	case NUM_EFFECTCLASSS:
	case EFFECTCLASS_NONE:
	    break;
	case EFFECTCLASS_HEAL:
	    descr.sprintf("Heals %d (%s)",
			    def->amount,
			    glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_GAINMANA:
	    descr.sprintf("Regains %d (%s) mana",
			    def->amount,
			    glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_LOSEMANA:
	    descr.sprintf("Drains %d (%s) mana",
			    def->amount,
			    glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_GAINFOOD:
	    descr.sprintf("Creates %d (%s) food",
			    def->amount,
			    glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_LOSEFOOD:
	    descr.sprintf("Consumes %d (%s) food",
			    def->amount,
			    glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_POISON:
	    descr.sprintf("Poisons for %d turns",
			    def->duration);
	    break;
	case EFFECTCLASS_CURE:
	    descr.sprintf("Cures poison");
	    break;
	case EFFECTCLASS_DAMAGE:
	    descr.sprintf("Causes %d %s damage (%s)",
			def->amount,
			GAMEDEF::elementdef(def->element)->name,
			glb_distributiondefs[def->distribution].name);
	    break;
	case EFFECTCLASS_RESIST:
	    descr.sprintf("Resist %s for %d turns",
			GAMEDEF::elementdef(def->element)->name,
			def->duration);
	    break;
	case EFFECTCLASS_VULNERABLE:
	    descr.sprintf("Vulnerable to %s for %d turns",
			GAMEDEF::elementdef(def->element)->name,
			def->duration);
	    break;
    }

    return descr;
}

bool
MOB::applyEffect(MOB *src, EFFECT_NAMES effect, ATTACKSTYLE_NAMES attackstyle)
{
    if (effect == EFFECT_NONE)
	return false;

    EFFECT_DEF	*def = GAMEDEF::effectdef(effect);

    switch ((EFFECTCLASS_NAMES) def->type)
    {
	case NUM_EFFECTCLASSS:
	case EFFECTCLASS_NONE:
	    return false;

	case EFFECTCLASS_HEAL:
	{
	    BUF		healstring;
	    int		heal = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    heal = MIN(heal, getMaxHP()-getHP());

	    healstring.sprintf("%%S <heal> %d damage.", heal);
	    formatAndReport(healstring);
	    gainHP(heal);
	    break;
	}

	case EFFECTCLASS_GAINMANA:
	{
	    BUF		healstring;
	    int		heal = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    heal = MIN(heal, getMaxMP()-getMP());

	    healstring.sprintf("%%S <regain> %d mana.", heal);
	    formatAndReport(healstring);
	    gainMP(heal);
	    break;
	}

	case EFFECTCLASS_LOSEMANA:
	{
	    BUF		healstring;
	    int		heal = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    heal = MIN(heal, getMP());

	    healstring.sprintf("%%S <be> drained of %d mana.", heal);
	    formatAndReport(healstring);
	    gainMP(-heal);
	    break;
	}

	case EFFECTCLASS_GAINFOOD:
	{
	    BUF		healstring;
	    int		heal = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    heal = MIN(heal, getMaxFood()-getFood());

	    healstring.sprintf("%%S <gain> %d food.", heal);
	    formatAndReport(healstring);
	    gainFood(heal);
	    break;
	}

	case EFFECTCLASS_LOSEFOOD:
	{
	    BUF		healstring;
	    int		heal = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    heal = MIN(heal, getFood());

	    healstring.sprintf("%%S <be> drained of %d food.", heal);
	    formatAndReport(healstring);
	    eatFood(heal);
	    break;
	}

	case EFFECTCLASS_POISON:
	{
	    ITEM		*item = ITEM::create(ITEM_POISON);

	    item->setTimer(def->duration);

	    addItem(item);
	    break;
	}

	case EFFECTCLASS_CURE:
	{
	    ITEM		*item = lookupItem(ITEM_POISON);

	    if (!item)
	    {
		formatAndReport("%S <feel> healthy.");
	    }
	    else
	    {
		removeItem(item);
	    }
	    break;
	}

	case EFFECTCLASS_DAMAGE:
	{
	    int		damage = evaluateDistribution(def->amount, (DISTRIBUTION_NAMES) def->distribution);

	    formatAndReport("%S <be> damaged.");

	    return applyDamage(src, damage, (ELEMENT_NAMES) def->element, attackstyle);
	}

	case EFFECTCLASS_RESIST:
	{
	    ITEM		*item = ITEM::create(ITEM_RESIST);

	    item->setTimer(def->duration);
	    item->setElement((ELEMENT_NAMES) def->element);

	    addItem(item);
	    break;
	}

	case EFFECTCLASS_VULNERABLE:
	{
	    ITEM		*item = ITEM::create(ITEM_VULNERABLE);

	    item->setTimer(def->duration);
	    item->setElement((ELEMENT_NAMES) def->element);

	    addItem(item);
	    break;
	}
    }
    return false;
}

bool
MOB::applyDamage(MOB *src, int hits, ELEMENT_NAMES element, ATTACKSTYLE_NAMES attackstyle)
{
    // Being hit isn't boring.
    clearBoredom();
    stopWaiting("%S <be> hit, so <end> %r rest.");

    if (hasItem(ITEM_INVULNERABLE))
	return false;

    if (element == ELEMENT_LIGHT && hasItem(ITEM_BLIND))
    {
	return false;
    }

    if (src && src->isAvatar() && src != this && !myAngryWithAvatar)
    {
	if (getAI() == AI_PEACEFUL)
	{
	    formatAndReport("%S <grow> angry with %o!", src);
	    myAngryWithAvatar = true;
	}
    }

    // Adjust damage down for armour.
    float		 multiplier = 1;
    int			 reduction = 0;

    ensureHitPoints();

    ensureDamageReduction();

    ensureSpecial();

    reduction = getDamageReduction();

    // No armour if the attack came from within.
    if (attackstyle == ATTACKSTYLE_INTERNAL)
	reduction = 0;
    
    // Check for resist/vulnerability
    if (element != ELEMENT_NONE)
    {
	if (defn().vulnerability == element)
	    multiplier *= 2;
	if (defn().resistance == element)
	    multiplier /= 2;

	// Check any items...
	for (int i = 0; i < myInventory.entries(); i++)
	{
	    ITEM	*item = myInventory(i);
	    if (item->getDefinition() == ITEM_RESIST)
	    {
		if (item->element() == element)
		    multiplier /= 2;
	    }
	    if (item->getDefinition() == ITEM_VULNERABLE)
	    {
		if (item->element() == element)
		    multiplier *= 2;
	    }
	}
    }

    if (multiplier != 1.0f)
    {
	float		fhits = hits * multiplier;

	hits = int(fhits);
	fhits -= hits;

	// Add the round off error.
	if (rand_double() < fhits)
	    hits++;
    }

    // Add in reduction
    reduction = rand_range(0, reduction);
    hits -= reduction;
    if (hits < 0)
	hits = 0;

    if (hits >= getHP())
    {
	// Ensure we are dead to alive()
	myHP = 0;
	myNumDeaths++;
	// Record for inverse monster memory.
	edefn().totalkillsever++;
	// Rather mean.
	myMP = 0;

	// Death!
	if (src)
	    src->formatAndReport("%S <kill> %O!", this);
	else
	    formatAndReport("%S <be> killed!");

	// If there is a source, and they are swallowed,
	// they are released.
	if (src && src->isSwallowed())
	{
	    src->setSwallowed(false);
	}
	
	// If we are the avatar, special stuff happens.
	if (isAvatar())
	{
	    // never update on this thread!
	    // msg_update();
	    // TODO: Pause something here?
	}
	else
	{
	    // Acquire any loot.
	    ensureLoot();

	    int		gold = rand_choice(getLevel()+1);
	    if (defn().loot >= 0)
		gold = rand_choice(defn().loot+1);
	    if (gold)
	    {
		gainGold(gold);
	    }

	    // Drop any stuff we have.
	    int i;
	    for (i = 0; i < myInventory.entries(); i++)
	    {
		if (myInventory(i)->defn().isflag)
		    delete myInventory(i);
		else
		    myInventory(i)->move(pos());
	    }
	    myInventory.clear();
	}

	if (src && src->isAvatar())
	{
	    // Award our score.

	}

	// Flash the screen
	pos().postEvent(EVENTTYPE_FOREBACK, ' ', ATTR_INVULNERABLE);

	// Note that avatar doesn't actually die...
	if (!isAvatar())
	{
	    // Anyone who witnessed this can yell.
	    if (src && src->isAvatar() && isFriends(src))
	    {
		MOBLIST		allmobs;

		pos().map()->getAllMobs(allmobs);

		for (int i = 0; i < allmobs.entries(); i++)
		{
		    if (allmobs(i)->pos().isFOV() && allmobs(i) != this &&
			!allmobs(i)->isAvatar() &&
			allmobs(i)->alive() &&
			allmobs(i)->isFriends(src))
		    {
			allmobs(i)->actionYell(YELL_MURDERER);
			allmobs(i)->giftItem(ITEM_ENRAGED);
			break;
		    }
		}
	    }

	    // Don't actually delete, but die.
	    MAP		*map = pos().map();

	    {
		ensureCorpseChance();

		if (rand_chance(defn().corpsechance))
		{
		    ITEM	*corpse = ITEM::create(ITEM_CORPSE);
		    corpse->setMobType(getDefinition());

		    corpse->move(pos());
		}
	    }

	    myPos.removeMob(this);
	    map->addDeadMob(this);
	    clearAllPos();
	    loseTempItems();
	}
	else
	{
	    // Make sure we drop our blind attribute..
	    loseTempItems();

	    // End any meditation
	    myMeditatePos = POS();

	    // No matter what, the source sees it (we may be meditating)
	    if (src)
		src->mySawVictory = true;

	    if (pos().isFOV())
	    {
		MOBLIST		allmobs;

		pos().map()->getAllMobs(allmobs);

		for (int i = 0; i < allmobs.entries(); i++)
		{
		    if (allmobs(i)->pos().isFOV())
			allmobs(i)->mySawVictory = true;
		}
	    }
	}
	return true;
    }

    // They lived.  They get the chance to yell.
    if (src && src->isAvatar() && isFriends(src) && !isAvatar())
    {
	// This is a free action.
	actionYell(YELL_MURDERER);
	giftItem(ITEM_ENRAGED);
    }

    // Flash that they are hit.
    if (hits)
	pos().postEvent(EVENTTYPE_FOREBACK, ' ', ATTR_HEALTH);

    myHP -= hits;
    return false;
}

VERB_PERSON
MOB::getPerson() const
{
    if (isAvatar())
	return VERB_YOU;

    return VERB_IT;
}

bool
MOB::isFriends(const MOB *other) const
{
    if (other == this)
	return true;

    if (hasItem(ITEM_ENRAGED) || other->hasItem(ITEM_ENRAGED))
	return false;
    
    if (isAvatar())
    {
	if (other->defn().isfriendly)
	    return true;
	else
	    return false;
    }

    // Only the avatar is evil!
    if (defn().isfriendly)
    {
	return true;
    }
    else
    {
	// Monsters hate the avtar!
	if (other->isAvatar())
	    return false;
    }
    return true;
}

AI_NAMES
MOB::getAI() const
{
    return (AI_NAMES) defn().ai;
}

void
MOB::reportSquare(POS t)
{
    if (!isAvatar())
	return;

    if (t.mob() && t.mob() != this)
    {
	formatAndReport("%S <see> %O.", t.mob());
    }
    if (t.item())
    {
	ITEMLIST	itemlist;

	t.allItems(itemlist);
	if (itemlist.entries())
	{
	    BUF		msg;
	    msg.strcpy("%S <see> ");
	    for (int i = 0; i < itemlist.entries(); i++)
	    {
		if (i)
		{
		    if (i == itemlist.entries()-1)
			msg.strcat(" and ");
		    else
			msg.strcat(", ");
		}

		msg.strcat(itemlist(i)->getArticleName());
	    }

	    msg.strcat(".");
	    formatAndReport(msg);
	}
    }

    if (t.defn().describe)
    {
	formatAndReport("%S <see> %O.", t.defn().legend);
    }
}

void
MOB::meditateMove(POS t)
{
    myMeditatePos = t;
    reportSquare(t);
}

bool
MOB::actionBump(int dx, int dy)
{
    MOB		*mob;
    POS		 t;

    // Stand in place.
    if (!dx && !dy)
	return true;

    if (isMeditating())
    {
	// We are free of our body!
	t = myMeditatePos.delta(dx, dy);
	if (t.defn().ispassable)
	{
	    meditateMove(t);
	    return true;
	}
	// Wall slide...
	if (dx && dy && isAvatar())
	{
	    // Try to wall slide, cause we are pretty real time here
	    // and it is frustrating to navigate curvy passages otherwise.
	    t = myMeditatePos.delta(dx, 0);
	    if (!rand_choice(2) && t.defn().ispassable)
	    { meditateMove(t); return true; }

	    t = myMeditatePos.delta(0, dy);
	    if (t.defn().ispassable)
	    { meditateMove(t); return true; }

	    t = myMeditatePos.delta(dx, 0);
	    if (t.defn().ispassable)
	    { meditateMove(t); return true; }
	}
	else if ((dx || dy) && isAvatar())
	{
	    // If we have
	    // ..
	    // @#
	    // ..
	    // Moving right we want to slide to a diagonal.
	    int		sdx, sdy;

	    // This bit of code is too clever for its own good!
	    sdx = !dx;
	    sdy = !dy;

	    t = myMeditatePos.delta(dx+sdx, dy+sdy);
	    if (!rand_choice(2) && t.defn().ispassable)
	    { meditateMove(t); return true; }

	    t = myMeditatePos.delta(dx-sdx, dy-sdy);
	    if (t.defn().ispassable)
	    { meditateMove(t); return true; }

	    t = myMeditatePos.delta(dx+sdx, dy+sdy);
	    if (t.defn().ispassable)
	    { meditateMove(t); return true; }
	}
	return false;
    }

    t = pos().delta(dx, dy);

    // If we are swallowed, we must attack.
    if (isSwallowed())
	return actionMelee(dx, dy);
    
    mob = t.mob();
    if (mob)
    {
	// Either kill or chat!
	if (mob->isFriends(this))
	{
	    if (isAvatar())
		return actionChat(dx, dy);
	}
	else
	{
	    return actionMelee(dx, dy);
	}
    }

    // No mob, see if we can move that way.
    // We let actionWalk deal with unable to move notification.
    return actionWalk(dx, dy);
}

bool
MOB::actionMeditate()
{
    if (isMeditating())
    {
	formatAndReport("%S <stop> meditating.");
	myMeditatePos = POS();
	// we don't want the user to be able to use up a fast
	// turn by doing this!
	PHASE_NAMES		phase;

	phase = map()->getPhase();
	if (phase == PHASE_FAST || phase == PHASE_QUICK)
	{
	    mySkipNextTurn = true;
	}
    }
    else
    {
	if (pos().tile() != TILE_MEDITATIONSPOT)
	{
	    formatAndReport("This location is not tranquil enough to support meditation.");
	    return false;
	}
	formatAndReport("%S <close> %r eyes and meditate.");
    }
    return true;
}

void
MOB::searchOffset(int dx, int dy, bool silent)
{
    POS		square = pos().delta(dx, dy);

    if (square.isTrap())
    {
	TRAP_NAMES		trap = (TRAP_NAMES) rand_choice(NUM_TRAPS);

	formatAndReport("%S find %O and disarm it.", glb_trapdefs[trap].name);
	square.clearTrap();

	square.postEvent((EVENTTYPE_NAMES)(EVENTTYPE_ALL | EVENTTYPE_LONG), 
			glb_trapdefs[trap].sym,
			(ATTR_NAMES) glb_trapdefs[trap].attr);
    }
    else if (square.tile() == TILE_SECRETDOOR)
    {
	formatAndReport("%S <reveal> a secret door.");
	square.setTile(TILE_DOOR);
	square.postEvent(EVENTTYPE_FOREBACK, '+', ATTR_SEARCH);
    }
    else
    {
	if (!silent)
	    square.postEvent(EVENTTYPE_FOREBACK, ' ', ATTR_SEARCH);
    }
}

bool
MOB::actionSearch(bool silent)
{
    mySearchPower = 1;
    if (!silent)
	formatAndReport("%S <search>.");

    int		ir, r;

    r = mySearchPower;
    // Note the < r to avoid double counting corners.
    for (ir = -r; ir < r; ir++)
    {
	searchOffset(ir, -r, silent);
	searchOffset(r, ir, silent);
	searchOffset(-ir, r, silent);
	searchOffset(-r, -ir, silent);
    }

    return true;
}

bool
MOB::actionDropButOne(ITEM *item)
{
    if (!item)
	return false;

    ITEM		*butone;

    butone = splitStack(item);
    assert(butone != item);

    if (butone == item)
    {
	addItem(butone);
	return false;
    }

    // Swap the two meanings
    butone->setStackCount(item->getStackCount());
    item->setStackCount(1);

    formatAndReport("%S <drop> %O.", butone);

    // And drop
    butone->move(pos());

    return true;
}

bool
MOB::actionDrop(ITEM *item)
{
    if (!item)
	return false;
    // Drop any stuff we have.
    int i;
    bool	fail = true;
    for (i = 0; i < myInventory.entries(); i++)
    {
	// Ignore special case items..
	// We don't want to drop "blindness" :>
	if (myInventory(i)->defn().isflag)
	    continue;

	if (myInventory(i) == item)
	{
	    formatAndReport("%S <drop> %O.", myInventory(i));
	    fail = false;
	    myInventory(i)->move(pos());
	    myInventory.set(i, 0);
	}
    }
    myInventory.collapse();

    updateEquippedItems();

    if (fail)
	formatAndReport("%S <drop> nothing.");

    return true;
}

bool
MOB::actionQuaff(ITEM *item)
{
    if (!item)
	return false;
    if (!item->isPotion())
    {
	formatAndReport("%S cannot drink %O.", item);
	return false;
    }

    item = splitStack(item);

    doEmote("*quaff*");
    formatAndReport("%S <quaff> %O.", item);
    bool		interesting = false;

    item->ensurePotionEffect();
    applyEffect(this, (EFFECT_NAMES)item->defn().potion_effect,
		    ATTACKSTYLE_INTERNAL);

    delete item;

    return true;
}

bool
MOB::actionThrow(ITEM *item, int dx, int dy)
{
    if (!item)
	return false;
    if (!item->isPotion())
    {
	msg_format("%S <be> not sufficiently aerodynamic.", item);
	return false;
    }

    item = splitStack(item);

    formatAndReport("%S <throw> %O.", item);
    bool		interesting = false;

    POTION_OP		op(this, item->getDefinition(), &interesting);

    u8			sym;
    ATTR_NAMES		attr;
    item->getLook(sym, attr);

    if (!doRangedAttack(5, 2, dx ,dy, '~', attr, "hit", op))
    {
	addItem(item);
	return false;
    }

    if (interesting)
    {
	item->markMagicClassKnown();
    }

    delete item;

    return true;
}

bool
MOB::actionEat(ITEM *item)
{
    if (!item)
	return false;

    if (!item->isFood())
    {
	formatAndReport("%S cannot eat %O.", item);
	return false;
    }

    if (getFood() >= getMaxFood())
    {
	formatAndReport("%S <be> too full to eat any more.");
	return false;
    }

    item = splitStack(item);

    formatAndReport("%S <eat> %O.", item);

    gainFood(100);

    item->ensureCorpseEffect();
    if (item->corpseEffect() != EFFECT_NONE)
    {
	applyEffect(this, item->corpseEffect(), ATTACKSTYLE_INTERNAL);
    }

    delete item;

    return true;
}

bool
MOB::actionBreak(ITEM *item)
{
    if (!item)
	return false;

    if (!item->isRanged())
    {
	formatAndReport("%S cannot break %O.", item);
	return false;
    }

    formatAndReport("%S <break> %O.", item);

    doEmote("*snap*");

    item->setBroken(true);

    return true;
}

bool
MOB::actionWear(ITEM *item)
{
    if (!item)
	return false;

    if (item->isRanged())
    {
	formatAndReport("You already know the best bow to use.");
	return false;
    }

    if (item->isMelee())
    {
	ITEM		*olditem = lookupWeapon();
	if (olditem)
	{
	    formatAndReport("%S <unwield> %O.", olditem);
	    olditem->setEquipped(false);
	}

	// If they are the same, the user wanted to toggle!
	if (olditem != item)
	{
	    formatAndReport("%S <wield> %O.", item);
	    item->setEquipped(true);
	}
	return true;
    }

    if (item->isArmour())
    {
	ITEM		*olditem = lookupArmour();
	if (olditem)
	{
	    formatAndReport("%S <take> off %O.", olditem);
	    olditem->setEquipped(false);
	}

	// If they are the same, the user wanted to toggle!
	if (olditem != item)
	{
	    formatAndReport("%S <put> on %O.", item);
	    item->setEquipped(true);
	}
	return true;
    }

    if (!item->isRing())
    {
	formatAndReport("%S cannot wear %O.", item);
	return false;
    }

    ITEM		*oldring = lookupRing();

    if (oldring)
    {
	formatAndReport("%S <take> off %O.", oldring);
	oldring->setEquipped(false);
    }

    // If they are the same, the user wanted to toggle!
    if (oldring != item)
    {
	formatAndReport("%S <put> on %O.", item);
	item->setEquipped(true);
    }

    return true;
}

const char *
getYellMessage(YELL_NAMES yell)
{
    const char *yellmsg;

    switch (yell)
    {
	case YELL_KEEPOUT:
	{
	    const char *msg[] =
	    {		// KEEPOUT
		"Keep out!",
		"Away!",
		"Leave!",
		"No Farther!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_MURDERER:
	{
	    const char *msg[] =
	    {		// MURDERER
		"Killer!",
		"Murderer!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_INVADER:
	{
	    const char *msg[] =
	    {		// INVADER
		"Invader!",
		"ALARM!",
		"ALARM!",
		"Code Yellow!",	// For yummy gold.
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_RANGED:
	{
	    const char *msg[] =
	    {		// RANGED
		"Far Threat!",
		"Code Bow!",	// Kobolds not that imaginative after all
		"Incoming!",
		"Archers!",	
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_KILL:
	{
	    const char *msg[] =
	    {		// KILL
		"Kill it now!",
		"Code Gold!",
		"Kill!",
		"Eviscerate It!",
		"Kill!",
		"Kill!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_KILLCHASE:
	{
	    const char *msg[] =
	    {		// KILL
		"No Mercy!",
		"Code Blood!",
		"Hunt It Down!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_LOCATION:
	{
	    const char *msg[] =
	    {		// LOCATION
		"It's here!",
		"Foe Sighted!",
		"Over here!",
		"I see it!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_HEARDLOCATION:
	{
	    const char *msg[] =
	    {		// LOCATION_HEARD 
		"It's there!",
		"Foe Located!",
		"Over there!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_TAUNT:
	{
	    const char *msg[] =
	    {		// LOCATION_HEARD 
		"Coward!",
		"You Flea!",
		"Mendicant!",
		"Wimp!",
		"Fool!",
		"Urchin!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
	case YELL_VICTORY:
	{
	    const char *msg[] =
	    {		// LOCATION_HEARD 
		"Victory!",
		"The @ is dead!",
		"Huzzah!",
		"Hooray!",
		"Gold!",
		0
	    };
	    yellmsg = rand_string(msg);
	    break;
	}
    }
    return yellmsg;
}

bool
MOB::actionYell(YELL_NAMES yell)
{
    const char	*yellflavour[] =
    {
	"%S <yell>: \"%o\"",
	"%S <shout>: \"%o\"",
	"%S <curse>: \"%o\"",
	0
    };
    const char *yellmsg = getYellMessage(yell);

    MOBLIST	hearmobs;

    pos().getAllConnectedRoomMobs(hearmobs);

    for (int i = 0; i < hearmobs.entries(); i++)
    {
	if (hearmobs(i)->isAvatar())
	{
	    // Don't use line of sight, but the same logic we use.
	    msg_format(rand_string(yellflavour), this, yellmsg);
	    doShout(yellmsg);
	}
	if (hearmobs(i) == this)
	    continue;
	hearmobs(i)->myHeardYell[yell] = true;
	if (hearmobs(i)->pos().roomId() == pos().roomId())
	{
	    // Extra flag to avoid spurious re-shouts
	    hearmobs(i)->myHeardYellSameRoom[yell] = true;
	}
    }

    return true;
}

bool
MOB::actionRotate(int angle)
{
    myPos = myPos.rotate(angle);
    if (isMeditating())
	myMeditatePos = myMeditatePos.rotate(angle);
    // This is always a free action.
    return false;
}

bool
MOB::actionClimb()
{
    TILE_NAMES		tile;

    tile = pos().tile();

    if (tile == TILE_DOWNSTAIRS)
    {
	// Climbing is interesting.
	myFleeCount = 0;
	clearBoredom();

	if (isAvatar())
	{
	    formatAndReport("%S <climb> down...");
	    // Chance to level up.
	    if (pos().map()->getDepth()+1 > getLevel())
	    {
		formatAndReport("%S <gain> experience!");
		gainLevel();
	    }
	    pos().map()->moveAvatarToDepth(pos().map()->getDepth()+1);
	    // Don't use time so you have a chance
	    // to move on new level.
	    return false;
	}
	else
	    return true;
    }
    else if (tile == TILE_UPSTAIRS)
    {
	if (isAvatar())
	{
	    if (pos().map()->getDepth() > 1)
	    {
		// Not done yet
		formatAndReport("%S <climb> up...");
		pos().map()->moveAvatarToDepth(pos().map()->getDepth()-1);
		// Don't use time so you have a chance
		// to move on new level.
		return false;
	    }
	    else
	    {
		if (!hasItem(ITEM_MACGUFFIN))
		{
		    formatAndReport("%S cannot climb without the %o!", GAMEDEF::itemdef(ITEM_MACGUFFIN)->name);
		    return false;
		}
		else
		{
		    formatAndReport("%S <escape>!");
		    myHasWon = true;
		    return false;
		}
	    }
	}
	else
	    return true;
    }

    formatAndReport("%S <see> nothing to climb here.");

    return false;
}

bool
MOB::actionChat(int dx, int dy)
{
    MOB		*victim;

    // This should never occur, but to avoid
    // embarassaments...
    if (!isAvatar())
	return false;
    
    victim = pos().delta(dx, dy).mob();
    if (!victim)
    {
	// Talk to self
	formatAndReport("%S <talk> to %O!", this);
	return false;
    }

    formatAndReport("%S <chat> with %O.", victim);
    
    return true;
}

bool
MOB::actionMelee(int dx, int dy)
{
    MOB		*victim;

    // If we are swallowed, attack the person on our square.
    if (isSwallowed())
	dx = dy = 0;

    POS		 t = pos().delta(dx, dy);

    victim = t.mob();
    if (!victim)
    {
	// Swing in air!
	msg_format("%S <swing> at empty air!", this);
	return false;
    }

    if (!victim->alive())
    {
	// The end GAme avatar, just silently ignore
	assert(!isAvatar());
	return false;
    }

    // Shooting is interesting.
    clearBoredom();
 
    int		damage;
    bool	victimdead;

    ensureMeleeAttack();

    damage = 0;
    if (evaluateMeleeWillHit(victim))
	damage = evaluateMeleeDamage();

    // attempt to kill victim
    if (damage)
    {
	msg_format("%S %v %O.", this, getMeleeVerb(),
				victim);
    }
    else
    {
	msg_format("%S <miss> %O.", this, victim);
	return true;
    }

    victimdead = victim->applyDamage(this, damage,
				    getMeleeElement(),
				    ATTACKSTYLE_MELEE);
    
    // Vampires regenerate!
    if (defn().isvampire)
    {
	myHP += damage;
    }

    // Grant our item...
    if (!victimdead)
    {
	if (attackdefn().effect != EFFECT_NONE)
	{
	    if (victim->applyEffect(this, (EFFECT_NAMES) attackdefn().effect,
				    ATTACKSTYLE_MELEE))
		return true;
	}

	// Swallow the victim
	// Seems silly to do this on a miss.
	// You also can't swallow people when you yourself are swallowed.
	if (!isSwallowed() && defn().swallows && damage)
	{
	    victim->setSwallowed(true);
	    victim->move(pos(), true);
	}

	// Steal something
	if (defn().isthief)
	{
	    ITEM		*item;
	    
	    // Get a random item from the victim
	    item = victim->getRandomItem();
	    if (item) 
	    {
		msg_format("%S <steal> %O.", this, item);
		// Theft successful.
		myFleeCount += 30;
		victim->removeItem(item, true);
		addItem(item);
	    }
	}
    }
	
    return true;
}

ATTACK_NAMES
MOB::getMeleeAttack() const
{
    ITEM	*w;

    w = lookupWeapon();
    if (!w)
	return (ATTACK_NAMES) defn().melee_attack;

    return (ATTACK_NAMES) w->defn().melee_attack;
}

const char *
MOB::getMeleeVerb() const
{
    ATTACK_NAMES	attack = getMeleeAttack();

    return GAMEDEF::attackdef(attack)->verb;
}

ELEMENT_NAMES
MOB::getMeleeElement() const
{
    ATTACK_NAMES	attack = getMeleeAttack();

    return (ELEMENT_NAMES) GAMEDEF::attackdef(attack)->element;
}

const char *
MOB::getMeleeWeaponName() const
{
    ITEM	*w;

    w = lookupWeapon();
    if (!w)
	return attackdefn().noun;

    return w->defn().name;
}

int
MOB::evaluateMeleeDamage() const
{
    ATTACK_NAMES	attack = getMeleeAttack();

    return evaluateAttackDamage(attack);
}

bool
MOB::evaluateMeleeWillHit(MOB *victim) const
{
    ATTACK_NAMES	attack = getMeleeAttack();

    return evaluateWillHit(attack, victim);
}

bool
MOB::evaluateWillHit(ATTACK_NAMES attack, MOB *victim) const
{
    // We can always hit inanimate objects?
    if (!victim)
	return true;

    ATTACK_DEF	*def = GAMEDEF::attackdef(attack);

    int		tohit = 10;

    tohit -= def->chancebonus;
    tohit += victim->defn().dodgebonus;

    tohit -= getLevel();
    tohit += victim->getLevel();

    // I keep turning back to the d20 system.  The simple fact is
    // that 5% is right at the human threshold for noticing, so is
    // by far the most convenient to express our boni with.
    int		dieroll = rand_range(1, 20);

    if (dieroll != 1 &&
	(dieroll == 20 || (dieroll >= tohit)))
	return true;
    return false;
}

int
MOB::evaluateAttackDamage(ATTACK_NAMES attack) const
{
    ATTACK_DEF	*def = GAMEDEF::attackdef(attack);

    int		base = def->damage;

    base = evaluateDistribution(base, (DISTRIBUTION_NAMES)def->distribution);

    return base;
}

int
MOB::evaluateDistribution(int amount, DISTRIBUTION_NAMES distribution)
{
    int		base = amount;

    switch (distribution)
    {
	case NUM_DISTRIBUTIONS:
	    break;
	case DISTRIBUTION_CONSTANT:
	    // damage is just damage
	    break;
	case DISTRIBUTION_UNIFORM:
	    // Uniform in double range gives same average.
	    base = rand_range(0, base * 2-1);
	    base++;
	    break;
	case DISTRIBUTION_BIMODAL:
	{
	    // Half the time we do a low-end distribution,
	    // half the time a high end.
	    if (rand_choice(2))
	    {
		// We don't double here, but add 50 percent
		// so our average matches.
		base += base / 2;
	    }
	    else
	    {
		base = base/2;
		if (!base)
		    base = 1;
	    }
		
	    //FALL THROUGH
	}
	case DISTRIBUTION_GAUSSIAN:
	{
	    // Sum of four uniforms.
	    int		total = 0;
	    total += rand_range(0, base * 2-1);
	    total += rand_range(0, base * 2-1);
	    total += rand_range(0, base * 2-1);
	    total += rand_range(0, base * 2-1);

	    total += 2;
	    total /= 4;

	    base = total + 1;
	    break;
	}
    }

    return base;
}

int
MOB::getRangedDamage() const
{
    return evaluateAttackDamage((ATTACK_NAMES)defn().range_attack);
}

int
MOB::getRangedRange() const
{
    int		range;

    range = defn().range_range;

    return range;
}

const char *
MOB::getRangedWeaponName() const
{
    return GAMEDEF::attackdef(defn().range_attack)->noun;
}

ELEMENT_NAMES
MOB::getRangedElement() const
{
    return (ELEMENT_NAMES) GAMEDEF::attackdef(defn().range_attack)->element;
}

void
MOB::getRangedLook(u8 &symbol, ATTR_NAMES &attr) const
{
    symbol = defn().range_symbol;
    attr = (ATTR_NAMES) defn().range_attr;
}


template <typename OP>
bool
MOB::doRangedAttack(int range, int area, int dx, int dy, 
		u8 symbol, ATTR_NAMES attr,
		const char *verb, OP op)
{
    int		rangeleft;
    MOB		*victim;

    // Check for friendly kill.
    victim = pos().traceBullet(range, dx, dy, &rangeleft);

    // Friendly fire is cool!

    // Shooting is interesting.
    clearBoredom();
 
    pos().displayBullet(range,
			dx, dy,
			symbol, attr,
			true);

    if (!victim)
    {
	// Shoot at air!
	// But, really, the fireball should explode!
	POS		vpos;
	vpos = pos().traceBulletPos(range, dx, dy, true);

	// Apply damage to everyone in range.
	vpos.fireball(this, area, symbol, attr, op); 

	return true;
    }

    // Attemp to kill victim
    msg_format("%S %v %O.", this, verb, victim);

    // NOTE: Bad idea to call a function on victim->pos() as
    // that likely will be deleted :>
    POS		vpos = victim->pos();

    // Apply damage to everyone in range.
    vpos.fireball(this, area, symbol, attr, op); 
    
    // The following code will keep the flame going past the target
#if 0
    while (isAvatar() && rangeleft && area == 1)
    {
	// Ensure our dx/dy is copacetic.
	vpos.setAngle(pos().angle());
	victim = vpos.traceBullet(rangeleft, dx, dy, &rangeleft);

	if (victim)
	{
	    msg_format("%S %v %O.", this, defn().range_verb,
				    victim);
	    vpos = victim->pos();
	    vpos.fireball(this, area, getRangedDPDF(), symbol, attr); 
	}
	else
	    rangeleft = 0;
    }
#endif

    return true;
}

bool
MOB::actionFire(int dx, int dy)
{
    // Check for no ranged weapon.
    if (!defn().range_valid && !lookupWand())
    {
	msg_format("%S <lack> a ranged attack!", this);
	return false;
    }


    ITEM_NAMES		ammo = ITEM_NONE;

    if (isAvatar() && lookupWand())
	ammo = (ITEM_NAMES) lookupWand()->defn().ammo;

    if (ammo != ITEM_NONE && !hasItem(ammo))
    {
	BUF		msg, ammoname;

	ammoname = gram_makeplural(ITEM::defn(ammo).name);
	msg.sprintf("%%S <be> out of %s!", ammoname.buffer());
	formatAndReport(msg);
	return false;
    }

    // No suicide.
    if (!dx && !dy)
    {
	msg_format("%S <decide> not to aim at %O.", this, this);
	return false;
    }
    
    // If swallowed, rather useless.
    if (isSwallowed())
    {
	msg_format("%S <do> not have enough room inside here.", this);
	return false;
    }

    u8		symbol;
    ATTR_NAMES	attr;
    getRangedLook(symbol, attr);

    // Use up an arrow!
    if (ammo != ITEM_NONE)
    {
	ITEM	*arrow;
	arrow = lookupItem(ammo);
	if (arrow)
	{
	    arrow->decStackCount();
	    if (!arrow->getStackCount())
	    {
		BUF		ammoname;

		ammoname = gram_makeplural(ITEM::defn(ammo).name);

		removeItem(arrow, true);
		formatAndReport("%S <shoot> the last of %r %o.", ammoname);
		delete arrow;
	    }
	}
    }

    ATTACK_OP	op(this, (ATTACK_NAMES) defn().range_attack);

    myRangeTimeout += defn().range_recharge;

    return doRangedAttack(getRangedRange(), 1, dx, dy, 
		    symbol, attr,
		    GAMEDEF::attackdef(defn().range_attack)->verb, op);
}

void
MOB::triggerManaUse(SPELL_NAMES spell, int manacost)
{
}

void
MOB::gainLevel()
{
    int		newhp = expectedHPforLevel(myLevel+1) - expectedHPforLevel(myLevel);
    int		newmp = expectedMPforLevel(myLevel+1) - expectedMPforLevel(myLevel);

    myLevel++;

    gainHP(newhp);
    gainMP(newmp);
}

int
MOB::getGold() const
{
    ITEM		*gold;
    int			 ngold;

    gold = lookupItem(ITEM_GOLD);
    
    ngold = 0;
    if (gold)
	ngold = gold->getStackCount();

    return ngold;
}

void
MOB::gainGold(int delta) 
{
    ITEM		*gold;

    if (!delta)
	return;

    gold = lookupItem(ITEM_GOLD);
    if (!gold && delta>0)
    {
	gold = ITEM::create(ITEM_GOLD);
	gold->setStackCount(0);
	addItem(gold);
    }
    if (!gold && delta<=0)
    {
	// Can't go negative?
	assert(!"Negative gold?");
	return;
    }
    
    gold->setStackCount(gold->getStackCount() + delta);
}

class SPELL_OP
{
public:
    SPELL_OP(MOB *src, SPELL_NAMES spell) 
    { mySrc = src; mySpell = spell; }

    void operator()(POS p)
    {
	MOB		*mob = p.mob();

	if (mob)
	{
	    mob->applyEffect(mySrc, 
		(EFFECT_NAMES) GAMEDEF::spelldef(mySpell)->effect,
		ATTACKSTYLE_RANGE);
	}
    }

private:
    SPELL_NAMES	 	mySpell;
    MOB			*mySrc;
};

bool
MOB::actionCast(SPELL_NAMES spell, int dx, int dy)
{
    // Pay the piper...
    int			manacost = 0, hpcost = 0;

    ensureSpell(spell);

    manacost = GAMEDEF::spelldef(spell)->mana;
    
    if (manacost > getMP())
    {
	// Don't let people suicide.
	formatAndReport("Without more mana, %S cannot cast %O!",
			GAMEDEF::spelldef(spell)->name);
	msg_newturn();
	return false;
    }

    if (hpcost)
    {
	formatAndReport("Paying with %r own life, %S <cast> %O.",
			    GAMEDEF::spelldef(spell)->name);
    }
    else
    {
	// Less dramatic.
	formatAndReport("%S <cast> %O.",
			    GAMEDEF::spelldef(spell)->name);
    }

    if (applyDamage(this, hpcost, ELEMENT_NONE, ATTACKSTYLE_INTERNAL))
	return true;

    gainMP(-manacost);
    
    triggerManaUse(spell, manacost);

    // Fire ball self or in a direction...
    SPELL_OP		op(this, spell);

    doRangedAttack(10, GAMEDEF::spelldef(spell)->radius,
		    dx, dy,
		    GAMEDEF::spelldef(spell)->symbol,
		    (ATTR_NAMES) GAMEDEF::spelldef(spell)->attr,
		    GAMEDEF::spelldef(spell)->verb,
		    op);
    return true;
}

bool
MOB::actionWalk(int dx, int dy)
{
    MOB		*victim;
    POS		 t;

    // If we are swallowed, we can't move.
    if (isSwallowed())
	return false;
    
    t = pos().delta(dx, dy);
    
    victim = t.mob();
    if (victim)
    {
	// If we are the avatar, screw being nice!
	if (isAvatar() && !isFriends(victim))
	    return actionMelee(dx, dy);
	// If it is thet avatar we are bumping into, screw being
	// nice.
	if (victim->isAvatar() && !isFriends(victim))
	    return actionMelee(dx, dy);

	// formatAndReport("%S <bump> into %O.", victim);
	// Bump into the mob.

	// Nah, tell the map we want a delayed move.
	// Do not delay move the avatar, however!
	if (victim->isAvatar() || isAvatar())
	    return false;

	return pos().map()->requestDelayedMove(this, dx, dy);
    }

    // Check to see if we can move that way.
    if (canMove(t))
    {
	TILE_NAMES	tile;
	
	// Determine if it is a special move..
	tile = t.tile();
	if (t.defn().isdiggable &&
	    !t.defn().ispassable &&	
	    defn().candig)
	{
	    return t.digSquare();
	}

	POS		opos = pos();
	
	// Move!
	move(t);

	// If we are a breeder, chance to reproduce!
	if (defn().breeder)
	{
	    static  int		lastbreedtime = -10;

	    // Only breed every 7 phases at most so one can
	    // always kill off the breeders.
	    // Only breed while the player is in the dungeon
	    bool		 canbreed = false;

	    if (getAvatar() && getAvatar()->alive() &&
		getAvatar()->pos().roomType() != ROOMTYPE_VILLAGE)
	    {
		canbreed = true;
	    }

	    // Hard cap
	    int		mobcount = pos().map()->getMobCount(getDefinition());
	    if (mobcount > 50)
		canbreed = false;

	    // Every new breeder increases the chance of breeding
	    // by 1 roll.  Thus we need at least the base * mobcount
	    // to just keep the breeding chance linear.
	    if (canbreed && 
		(lastbreedtime < map()->getTime() - 12) &&
		!rand_choice(3 + 3*mobcount))
	    {
		MOB		*child;
		
		formatAndReport("%S <divide>!");
		child = MOB::create(getDefinition());
		child->move(opos);
		lastbreedtime = map()->getTime();
	    }
	}
	
	// If we are the avatar and something is here, pick it up.
	if (isAvatar() && t.item())
	    actionPickup();

	// Avatars set off traps because they are clumsy
	bool		hadtrap = false;
	if (isAvatar() && t.isTrap())
	{
	    hadtrap = true;
	    t.clearTrap();
	    TRAP_NAMES		trap = (TRAP_NAMES) rand_choice(NUM_TRAPS);

	    t.postEvent((EVENTTYPE_NAMES)(EVENTTYPE_ALL | EVENTTYPE_LONG), 
			    glb_trapdefs[trap].sym,
			    (ATTR_NAMES) glb_trapdefs[trap].attr);
	    DPDF	dpdf(0);

	    formatAndReport("%S <set> off %O!", glb_trapdefs[trap].name);

	    if (glb_trapdefs[trap].item != ITEM_NONE)
	    {
		ITEM		*item = ITEM::create((ITEM_NAMES)glb_trapdefs[trap].item);
		item->setStackCount(1);
		addItem(item);
	    }

	    applyDamage(0, dpdf.evaluate(), 
			    (ELEMENT_NAMES) glb_trapdefs[trap].element,
			    ATTACKSTYLE_MELEE);
	}

	// Apply searching

	// Attempt to climb!
	if (alive() && isAvatar() && t.tile() == TILE_DOWNSTAIRS)
	{
	    msg_report("Press > to climb down.");
	}
	// Attempt to climb!
	if (alive() && isAvatar() && t.tile() == TILE_UPSTAIRS)
	{
	    msg_report("Press < to climb up.");
	}

	return true;
    }
    else
    {
	if (dx && dy && isAvatar())
	{
	    // Try to wall slide, cause we are pretty real time here
	    // and it is frustrating to navigate curvy passages otherwise.
	    if (!rand_choice(2) && canMoveDir(dx, 0, false))
		if (actionWalk(dx, 0))
		    return true;
	    if (canMoveDir(0, dy, false))
		if (actionWalk(0, dy))
		    return true;
	    if (canMoveDir(dx, 0, false))
		if (actionWalk(dx, 0))
		    return true;
	}
	else if ((dx || dy) && isAvatar())
	{
	    // If we have
	    // ..
	    // @#
	    // ..
	    // Moving right we want to slide to a diagonal.
	    int		sdx, sdy;

	    // This bit of code is too clever for its own good!
	    sdx = !dx;
	    sdy = !dy;

	    if (!rand_choice(2) && canMoveDir(dx+sdx, dy+sdy, false))
		if (actionWalk(dx+sdx, dy+sdy))
		    return true;
	    if (canMoveDir(dx-sdx, dy-sdy, false))
		if (actionWalk(dx-sdx, dy-sdy))
		    return true;
	    if (canMoveDir(dx+sdx, dy+sdy, false))
		if (actionWalk(dx+sdx, dy+sdy))
		    return true;
	    
	}

	formatAndReport("%S <be> blocked by %O.", t.defn().legend);
	if (isAvatar())
	    msg_newturn();
	// Bump into a wall.
	return false;
    }
}

bool
MOB::actionPickup()
{
    ITEM		*item;

    item = pos().item();

    return actionPickup(item);
}

bool
MOB::actionPickup(ITEM *item)
{
    if (!item)
    {
	formatAndReport("%S <grope> the ground foolishly.");
	return false;
    }

    if (item->pos() != pos())
    {
	formatAndReport("%S <be> too far away to pick up %O.", item);
	return false;
    }

    item->ensureNamed();

    // Pick up the item!
    formatAndReport("%S <pick> up %O.", item);

    // No longer coveted!
    item->setInterestedUID(INVALID_UID);
    item->move(POS());

    if (item->defn().discardextra && hasItem(item->getDefinition()))
    {
	// Replace broken items...
	ITEM		*olditem = lookupItem(item->getDefinition());
	
	if (olditem->isBroken() && !item->isBroken())
	{
	    formatAndReport("%S <replaces> %O.", olditem);
	    olditem->setBroken(false);
	}

	formatAndReport("%S already <have> one of these, so %S <discard> it.");

	delete item;
    }
    else
	addItem(item);

    return true;
}

void
MOB::updateEquippedItems()
{
    ITEM		*item;

    // Non avatars auto equip.
    if (!isAvatar())
    {
	// Mark everything unequipped
	for (int i = 0; i < myInventory.entries(); i++)
	{
	    item = myInventory(i);

	    if (!item->isRing())
		item->setEquipped(false);
	}

	// Equip our weapon and armour
	ITEM		*weap = 0, *arm = 0, *wand = 0;

	for (int i = 0; i < myInventory.entries(); i++)
	{
	    if (myInventory(i)->isMelee() && !myInventory(i)->isBroken())
		weap = aiLikeMoreWeapon(myInventory(i), weap);
	}
	if (weap)
	    weap->setEquipped(true);
	for (int i = 0; i < myInventory.entries(); i++)
	{
	    if (myInventory(i)->isRanged() && !myInventory(i)->isBroken())
		wand = aiLikeMoreWand(myInventory(i), wand);
	}
	if (wand)
	    wand->setEquipped(true);
	for (int i = 0; i < myInventory.entries(); i++)
	{
	    if (myInventory(i)->isArmour() && !myInventory(i)->isBroken())
		arm = aiLikeMoreArmour(myInventory(i), arm);
	}
	if (arm)
	    arm->setEquipped(true);
    }

    // Count total items
    int		total = 0;
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->defn().isflag)
	    continue;
	total++;
    }
    if (total > 15)
    {
	// Yay!  Chacne to use count down operator.
	// We want to run in reverse here.
	formatAndReport("%r backpack is out of room!");
	for (int i = myInventory.entries(); i --> 0; )
	{
	    item = myInventory(i);
	    if (item->defn().isflag)
		continue;

	    formatAndReport("%S <drop> %O.", item);
	    item->move(pos());
	    myInventory.removeAt(i);

	    total--;
	    if (total <= 15)
		break;
	}
    }
}

void
MOB::addItem(ITEM *item)
{
    int			 i;

    if (!item)
	return;
    
    // Alert the world to our acquirement.
    if (isAvatar() || pos().isFOV())
	if (item->defn().gaintxt)
	    msg_format(item->defn().gaintxt,
			this, item);

    // First, check to see if we can merge...
    for (i = 0; i < myInventory.entries(); i++)
    {
	if (item->canStackWith(myInventory(i)))
	{
	    myInventory(i)->combineItem(item);
	    delete item;
	    return;
	}
    }

    // Brand new item.
    myInventory.append(item);

    // If we didn't stack, we may need to update our best items.
    updateEquippedItems();
}

ITEM *
MOB::splitStack(ITEM *item)
{
    if (item->getStackCount() > 1)
    {
	ITEM	*result;

	item->decStackCount();
	result = item->createCopy();
	result->setStackCount(1);

	return result;
    }
    else
    {
	removeItem(item);
	return item;
    }
}

void
MOB::removeItem(ITEM *item, bool quiet)
{
    // Alert the world to our acquirement.
    if ((isAvatar() || pos().isFOV()) && !quiet)
	if (item->defn().losetxt)
	    msg_format(item->defn().losetxt,
			this, item);

    myInventory.removePtr(item);

    updateEquippedItems();
}

void
MOB::loseAllItems()
{
    int		i;
    ITEM	*item;

    for (i = myInventory.entries(); i --> 0;)
    {
	item = myInventory(i);
	removeItem(item, true);
	delete item;
    }
}

void
MOB::loseTempItems()
{
    int		i;
    ITEM	*item;

    for (i = myInventory.entries(); i --> 0;)
    {
	item = myInventory(i);
	// All flags and timers are temp.
	if (item->getTimer() >= 0 || item->defn().isflag)
	{
	    removeItem(item, true);
	    delete item;
	}
    }
}

void
MOB::save(ostream &os) const
{
    int		val;
    u8		c;

    val = getDefinition();
    os.write((const char *) &val, sizeof(int));

    myName.save(os);

    myPos.save(os);
    myMeditatePos.save(os);
    myTarget.save(os);
    myHome.save(os);

    c = myWaiting;
    os.write((const char *) &c, sizeof(c));

    os.write((const char *) &myLevel, sizeof(int));
    os.write((const char *) &myHP, sizeof(int));
    os.write((const char *) &myMP, sizeof(int));
    os.write((const char *) &myFood, sizeof(int));
    os.write((const char *) &myAIState, sizeof(int));
    os.write((const char *) &myFleeCount, sizeof(int));
    os.write((const char *) &myBoredom, sizeof(int));
    os.write((const char *) &myYellHystersis, sizeof(int));
    os.write((const char *) &myNumDeaths, sizeof(int));
    os.write((const char *) &myUID, sizeof(int));
    os.write((const char *) &mySearchPower, sizeof(int));
    os.write((const char *) &myCowardice, sizeof(int));
    os.write((const char *) &myKills, sizeof(int));
    os.write((const char *) &myRangeTimeout, sizeof(int));

    val = isSwallowed();
    os.write((const char *) &val, sizeof(int));

    val = myAngryWithAvatar;
    os.write((const char *) &val, sizeof(int));

    int			 numitem;
    int			 i;

    c = mySawMurder;
    os.write((const char *) &c, 1);
    c = mySawMeanMurder;
    os.write((const char *) &c, 1);
    c = mySawVictory;
    os.write((const char *) &c, 1);
    c = myAvatarHasRanged;
    os.write((const char *) &c, 1);

    YELL_NAMES		yell;
    FOREACH_YELL(yell)
    {
	c = myHeardYell[yell];
	os.write((const char *) &c, 1);
	c = myHeardYellSameRoom[yell];
	os.write((const char *) &c, 1);
    }

    numitem = myInventory.entries();
    os.write((const char *) &numitem, sizeof(int));

    for (i = 0; i < myInventory.entries(); i++)
    {
	myInventory(i)->save(os);
    }
}

MOB *
MOB::load(istream &is)
{
    int		 val, num, i;
    u8		 c;
    MOB		*mob;

    mob = new MOB();

    is.read((char *)&val, sizeof(int));
    mob->myDefinition = (MOB_NAMES) val;

    if (mob->myDefinition >= GAMEDEF::getNumMob())
    {
	mob->myDefinition = MOB_NONE;
    }

    mob->myName.load(is);

    mob->myPos.load(is);
    mob->myMeditatePos.load(is);
    mob->myTarget.load(is);
    mob->myHome.load(is);

    is.read((char *) &c, sizeof(c));
    mob->myWaiting = c ? true : false;
    
    is.read((char *)&mob->myLevel, sizeof(int));
    is.read((char *)&mob->myHP, sizeof(int));
    is.read((char *)&mob->myMP, sizeof(int));
    is.read((char *)&mob->myFood, sizeof(int));
    is.read((char *)&mob->myAIState, sizeof(int));
    is.read((char *)&mob->myFleeCount, sizeof(int));
    is.read((char *)&mob->myBoredom, sizeof(int));
    is.read((char *)&mob->myYellHystersis, sizeof(int));
    is.read((char *)&mob->myNumDeaths, sizeof(int));
    is.read((char *)&mob->myUID, sizeof(int));
    is.read((char *)&mob->mySearchPower, sizeof(int));
    is.read((char *)&mob->myCowardice, sizeof(int));
    is.read((char *)&mob->myKills, sizeof(int));
    is.read((char *)&mob->myRangeTimeout, sizeof(int));
    glb_reportUID(mob->myUID);

    is.read((char *)&val, sizeof(int));
    mob->setSwallowed(val ? true : false);

    is.read((char *)&val, sizeof(int));
    mob->myAngryWithAvatar = (val ? true : false);

    is.read((char *)&c, 1);
    mob->mySawMurder = (c ? true : false);
    is.read((char *)&c, 1);
    mob->mySawMeanMurder = (c ? true : false);
    is.read((char *)&c, 1);
    mob->mySawVictory = (c ? true : false);
    is.read((char *)&c, 1);
    mob->myAvatarHasRanged = (c ? true : false);
    
    YELL_NAMES		yell;
    FOREACH_YELL(yell)
    {
	is.read((char *)&c, 1);
	mob->myHeardYell[yell] = (c ? true : false);
	is.read((char *)&c, 1);
	mob->myHeardYellSameRoom[yell] = (c ? true : false);
    }

    is.read((char *)&num, sizeof(int));

    for (i = 0; i < num; i++)
    {
	mob->myInventory.append(ITEM::load(is));
    }

    return mob;
}

ITEM *
MOB::getItemFromId(int itemid) const
{
    for (int i = 0; i < myInventory.entries(); i++) 
    { 
	if (myInventory(i)->getUID() == itemid) 
	    return myInventory(i);
    } 
    return 0; 
}

bool
MOB::hasVisibleEnemies() const
{
    PTRLIST<MOB *> list;

    getVisibleEnemies(list);
    if (list.entries())
	return true;
    return false;
}

bool
MOB::actionDropSurplus()
{
    ITEM		*weap, *wand;

    weap = lookupWeapon();
    wand = lookupWand();
    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i) == weap)
	    continue;
	if (myInventory(i) == wand)
	    continue;
	if (myInventory(i)->isPotion())
	    continue;

	if (myInventory(i)->defn().isflag)
	    continue;

	return actionDrop(myInventory(i));
    }
    // Check if our wand/weap is redundant
    if (weap && weap->getStackCount() > 1)
    {
	return actionDropButOne(weap);
    }
    if (wand && wand->getStackCount() > 1)
    {
	return actionDropButOne(wand);
    }
    return false;
}

bool
MOB::hasSurplusItems() const
{
    bool	hasrange = false;
    bool	hasmelee = false;

    for (int i = 0; i < myInventory.entries(); i++)
    {
	// We can always drink these.
	if (myInventory(i)->isPotion())
	    continue;
	if (myInventory(i)->defn().isflag)
	    continue;
	// Gold isn't surplus!
	if (myInventory(i)->getDefinition() == ITEM_GOLD)
	    continue;
	// Stuff we wear is important!
	if (myInventory(i)->isEquipped())
	    continue;
	return true;
    }

    return false;
}

int
MOB::numSurplusRange() const
{
    int		numrange = 0;

    for (int i = 0; i < myInventory.entries(); i++)
    {
	if (myInventory(i)->isRanged())
	{
	    numrange += myInventory(i)->getStackCount();
	}
    }

    // Reserve one range for our own use.
    return numrange - 1;
}

void
MOB::getVisibleEnemies(PTRLIST<MOB *> &list) const
{
    // If we are blind, we can't see anyone.
    if (hasItem(ITEM_BLIND))
	return;

    // We only care about the non-avatar case now.
    if (isAvatar())
	return;

    // Check if we are in FOV.  If so, the avatar is visible.
    if (getAvatar())
    {
	// Ignore dead avatars.
	if (!getAvatar()->alive())
	{
	    return;
	}
	// Ignore none hostile
	if (isFriends(getAvatar()))
	    return;

	// We need the double check because if you are meditating
	// and put them in the fov, that doesn't mean that 
	// the avatar is now in fov.
	if (pos().isFOV() && getAvatar()->pos().isFOV())
	{
	    list.append(getAvatar());
	}
    }
}

void
MOB::formatAndReport(const char *msg)
{
    if (isAvatar() || pos().isFOV())
    {
	msg_format(msg, this);
    }
}

void
MOB::formatAndReport(const char *msg, MOB *object)
{
    if (isAvatar() || pos().isFOV())
    {
	msg_format(msg, this, object);
    }
}

void
MOB::formatAndReport(const char *msg, ITEM *object)
{
    if (isAvatar() || pos().isFOV())
    {
	msg_format(msg, this, object);
    }
}

void
MOB::formatAndReport(const char *msg, const char *object)
{
    if (isAvatar() || pos().isFOV())
    {
	msg_format(msg, this, object);
    }
}

int
MOB::numberMeleeHostiles() const
{
    int		dx, dy;
    MOB		*mob;
    int		hostiles = 0;

    FORALL_8DIR(dx, dy)
    {
	mob = pos().delta(dx, dy).mob();

	if (mob && !mob->isFriends(this))
	    hostiles++;
    }
    return hostiles;
}

void
MOB::clearCollision()
{
    myCollisionSources.clear();
    myCollisionTarget = 0;
}

void
MOB::setCollisionTarget(MOB *target)
{
    if (myCollisionTarget == target)
	return;

    assert(!target || !myCollisionTarget);

    myCollisionTarget = target;
    if (target)
	target->collisionSources().append(this);
}

bool
MOB::actionPortalFire(int dx, int dy, int portal)
{
    // No suicide.
    if (!dx && !dy)
    {
	msg_format("%S <decide> not to aim at %O.", this, this);
	return false;
    }
    
    // If swallowed, rather useless.
    if (isSwallowed())
    {
	msg_format("%S do not have enough room inside here.", this);
	return false;
    }

    clearBoredom();

    // Always display now.
 
    u8		symbol = '*';
    ATTR_NAMES	attr;
    int		portalrange = 60;

    attr = ATTR_BLUE;
    if (portal)
	attr = ATTR_ORANGE;
    
    pos().displayBullet(portalrange,
			dx, dy,
			symbol, attr,
			false);

    POS		vpos;
    vpos = pos().traceBulletPos(portalrange, dx, dy, false, false);
    
    if (vpos.tile() != TILE_WALL && vpos.tile() != TILE_PROTOPORTAL)
    {
	msg_format("The portal dissipates at %O.", 0, vpos.defn().legend);
	return false;
    }

    if (!glb_roomtypedefs[vpos.roomType()].allowportal)
    {
	formatAndReport("The walls of this area seem unable to hold portals.");
	return false;
    }


    // We want sloppy targeting for portals.
    // This means that given a portal location of # fired at from the south,
    // we want:
    //
    // 1#1
    // 2*2
    //
    // as potential portals.
    //
    // One fired from the south-west
    //  2
    // 1#2
    // *1

    POS		alt[5];

    if (dx && dy)
    {
	alt[0] = vpos;
	alt[1] = vpos.delta(-dx, 0);
	alt[2] = vpos.delta(0, -dy);
	alt[3] = vpos.delta(0, dy);
	alt[4] = vpos.delta(dx, 0);
    }
    else if (dx)
    {
	alt[0] = vpos;
	alt[1] = vpos.delta(0, 1);
	alt[2] = vpos.delta(0, -1);
	alt[3] = vpos.delta(-dx, 1);
	alt[4] = vpos.delta(-dx, -1);
    }
    else
    {
	alt[0] = vpos;
	alt[1] = vpos.delta(1, 0);
	alt[2] = vpos.delta(-1, 0);
	alt[3] = vpos.delta(1, -dy);
	alt[4] = vpos.delta(-1, -dy);
    }

    for (int i = 0; i < 5; i++)
    {
	if (buildPortalAtLocation(alt[i], portal))
	    return true;
    }
    msg_report("The wall proved too unstable to hold a portal.");

    return false;
}

bool
MOB::buildPortalAtLocation(POS vpos, int portal) const
{
    int		origangle;

    origangle = vpos.angle();

    // We want all our portals in the same space.
    vpos.setAngle(0);

    // Check if it is a valid portal pos?
    if (!vpos.prepSquareForDestruction())
    {
	return false;
    }

    // Verify the portal is well formed, ie, three neighbours are now
    // walls and the other is a floor.
    int		dir, floordir = -1;
    TILE_NAMES	tile;

    for (dir = 0; dir < 4; dir++)
    {
	tile = vpos.delta4Direction(dir).tile();
	if (tile == TILE_FLOOR)
	{
	    if (floordir < 0)
		floordir = dir;
	    else
	    {
		// Uh oh.
		return false;

	    }
	}
	else if (tile == TILE_WALL || tile == TILE_PROTOPORTAL)
	{
	    // All good.
	}
	else
	{
	    // Uh oh.
	    return false;
	}
    }

    // Failed to find a proper portal.
    if (floordir < 0)
	return false;

    // In floordir+2 we will be placing the mirror of the opposite
    // portal.  It is important that square is not accessible.  We
    // merely make sure it isn't a floor
    // Still an issue of having ants dig it out.  Ideally we'd
    // have the virtual portal flagged with MAPFLAG_PORTAL but
    // we don't do that currently in buildPortal and doing so
    // would mean we'd have to clean it up properly.

    POS		virtualportal;

    virtualportal = vpos.delta4Direction((floordir+2)&3);
    virtualportal = virtualportal.delta4Direction((floordir+2)&3);

    // We now point to the square behind the proposed virtual portal
    // this should be wall or invalid
    tile = virtualportal.tile();
    if (tile != TILE_WALL && tile != TILE_PROTOPORTAL && tile != TILE_INVALID)
	return false;
    // Try neighbours.
    tile = virtualportal.delta4Direction((floordir+1)&3).tile();
    if (tile != TILE_WALL && tile != TILE_PROTOPORTAL && tile != TILE_INVALID)
	return false;
    tile = virtualportal.delta4Direction((floordir-1)&3).tile();
    if (tile != TILE_WALL && tile != TILE_PROTOPORTAL && tile != TILE_INVALID)
	return false;

    // We now know we aren't an isolated island.  Yet.

    vpos.map()->buildUserPortal(vpos, portal, (floordir+2) & 3, origangle);

    return true;
}

